#! /usr/bin/env python
#

""" LightCurve Query from IRSA """



    
import requests
import warnings
from io import StringIO

import numpy as np
from pandas import read_csv
"""
These are not lightcurves generated from alert packets. 
These are from the matching the epochal catalogs. Totally independent of alerts. 
The variable star/AGN community will be most interested in these. The “seeds” used for matching are detections from reference image coadds.
 
"""

BASEURL = "https://irsa.ipac.caltech.edu/cgi-bin/ZTF/nph_light_curves?"

# ==================== #
#                      #
#  COLOR and MPL       #
#                      #
# ==================== #

FILTER_COLORS = ["C2","C3","C1"]
FILTER_CODE   = ["zg","zr","zi"]

    
# ==================== #
#                      #
#  Building Queries    #
#                      #
# ==================== #

EXISTING_QUERY_PARAMS = "ID CIRCLE POS BAND BANDNAME MAG NUM_OBS TIME BAD_CATFLAGS_MASK COLLECTION FORMAT".split()
FOUR_PARAMS      = ["POS"]
THREE_PARAMS     = ["CIRCLE"]
TWO_PARAMS       = ["MAG"]
TWO_OR_ONE_PARAM = ["BAND","TIME"]

def build_query(**kwargs):
    """ """
    for i,[k,v] in enumerate(kwargs.items()):
        if v is None:
            continue
        k = k.upper()
        if k not in EXISTING_QUERY_PARAMS:
            raise ValueError("%s is not a queriable parameter. These are: "%k+", ".join(EXISTING_QUERY_PARAMS))
        if k in FOUR_PARAMS and len(v)!=4:
            raise ValueError("%s must have four parameters."%k+"{}".format(v)+" given")
        if k in THREE_PARAMS and len(v)!=3:
            raise ValueError("%s must have three parameters."%k+"{}".format(v)+" given")
        if k in TWO_PARAMS and len(v)!=2:
            raise ValueError("%s must have two parameters."%k+"{}".format(v)+" given")
        if k in TWO_OR_ONE_PARAM and len(v)>2:
            raise ValueError("%s must have less than two parameters."%k+"{}".format(v)+" given")
        
    
        if i==0:
            query = ''
        else:
            query+= '&'
                
        if k=="ID":
            for j,v_ in enumerate(str(v).split(",")):
                if j>0:
                    query+= '&'
                query+=("%s="%k+"{}".format(v_)).replace("[",'').replace("]","").replace(",","").replace("'","")

        else:
            query+=("%s="%k+"{}".format(v)).replace("[",'').replace("]","").replace(",","").replace("'","")
            
    return query
    
def build_url(**kwargs):
    """ """
    return BASEURL + build_query(**kwargs)



class LCQuery( object ):
    
    def __init__(self):
        """ """

    def query_id(self, id,  cookies=None, auth=None, **kwargs):
        """ 
        id: [string/int]
            The id parameter value is the identifier of a ZTF object, and comma-separated combinations thereof (or list of).
            * Example: 
            >>> id=686103400067717
            >>> id='686103400067717,686103400106565'
            >>> id=[686103400067717,686103400106565]
            
        """
        if np.any([k.upper() in ["BAND", "BANDNAME", "NOBS_MIN","MAG"]  for k in kwargs.keys()]):
            raise ValueError("Parameters BAND, BANDNAME, NOBS_MIN, and MAG are compatible with POS and CIRCLE but not with ID.")
        
        return self.download_data(cookies=cookies,auth=auth,
                                ID=",".join(np.atleast_1d(id)), **kwargs)

    def query_position(self, ra, dec, radius_arcsec, pos="circle", 
                        bandname=None, mag=None,
                        cookies=None, auth=None,
                        **kwargs):
        """ 
        Parameters
        ----------
        ra, dec: [float, float]
            coordinates in degrees. 
            ra and dec are assumed to be in the ICRS system. 
            This parameter restricts ZTF objects of interest to the circle of `radius`
            with center determined by `ra` and `dec`.
            The valid range of ra is [0,180] and of dec is [-90,90]. 
            
        radius_arcsec: [float]
            distance in arcsec.
            For performance reasons, the valid range is limited to (0,600].
    
        pos: [string] -optional-
            The POS parameter value must consist of a shape described by ICRS coordinates 
            in decimal degrees. It identifies the shape which contains ZTF objects of interest. 
            *The only shape currently supported is circle.*
            avaible pos:
            - circle

        // query options

        bandname: [string] -optional-
            The bandname parameter identifies by filter-id the wavelength interval(s) to be 
            searched for data. Possible values are "g", "r", and "i", 
            respectively equivalent to "1", "2" and "3", and comma-separated combinations thereof.
            // Implemented as a filter on the "fid" column of the ZTF objects table.
            * Examples:
            >>> Find only G-band data, which covers the wavelength range 410nm—550nm:
                bandname=g
            >>> Find only G-band and I-band data, which cover the wavelength ranges 410nm—550nm 
               and 700nm—900nm respectively:
               bandname=g,i

        mag: [1 or 2-value array] -optional-
            The mag parameter specifies a range in which the magnitude of ZTF objects of interest 
            must lie.
            // Implemented as a filter on the "medianmag" column of the ZTF objects table.
            * Examples:
            >>> mag=[17.0,17.7]

        **kwargs goes to download_data (e.g. num_obs, time, band, collection)

        Returns
        -------
        LCQuery
        """
        radius = radius_arcsec/3600
        return self.download_data(cookies=cookies,auth=auth,
                                POS=("{pos} {ra} {dec} {radius}".format(**locals())).split(),
                                bandname=bandname, mag=mag, **kwargs)
        
    @staticmethod
    def download_data(cookies=None, auth=None, **kwargs):
        """ 

        Parameters
        ----------
        
        All the following parameters could be given as kwargs:
        see https://irsa.ipac.caltech.edu/docs/program_interface/ztf_lightcurve_api.html
        
        
        ID: [string/int]
            The ID parameter value is the identifier of a ZTF object, and comma-separated combinations thereof.
            * Example: 
            >>> ID=686103400067717
            >>> ID=686103400067717,686103400106565
        
        CIRCLE: [3-values array]
            The CIRCLE parameter value consists of 3 elements, 
            all measured in degrees: RA (right ascension), DEC (declination), and RADIUS, 
            in that order. RA and DEC are assumed to be in the ICRS system. 
            As with other multi-element values, the elements must be separated by a single space. 
            This parameter restricts ZTF objects of interest to the circle of radius RADIUS 
            with center determined by RA and DEC.
            The valid range of RA is [0,180] and of DEC is [-90,90]. 
            For performance reasons, the valid range of RADIUS 
            is limited to (0,0.1667].
            * Example:
            >>> CIRCLE=-164.7 -5.8 0.1

        POS: [4-values array]
            The POS parameter value must consist of a shape described by ICRS coordinates 
            in decimal degrees. It identifies the shape which contains ZTF objects of interest. 
            The only shape currently supported is CIRCLE. 
            The three following elements of the value correspond to RA, DEC, and RADIUS 
            respectively; see the description of CIRCLE above.
            * Example:
            >>> POS=circle -164.7 -5.8 0.1

        BAND: [1 or 2-value array]
            The BAND parameter defines the wavelength interval, measured in meters, 
            to be searched for data. This interval is unbounded by default. 
            If (semi-)bounded, the interval includes the bounding value(s). 
            A BAND constraint is satisfied if the interval intersects the wavelength coverage 
            of the observation.
            * Examples:
            >>> Retrieve only data in the wavelength range 410nm—550nm:
                BAND=4.10e-7 5.50e-7
            >>> Retrieve data with wavelength no less than 1.4 micron:
                BAND=1.4e-6 Inf
            >>> Retrieve data with wavelength no more than 1.4 micron:
                BAND=0 1.4e-6
            >>> Retrieve data that includes 2.2 micron:
                BAND=2.2e-6

        BANDNAME: [string] // could combine several
            The BANDNAME parameter identifies by filter-id the wavelength interval(s) to be 
            searched for data. Possible values are "g", "r", and "i", 
            respectively equivalent to "1", "2" and "3", and comma-separated combinations thereof.
            // Implemented as a filter on the "fid" column of the ZTF objects table.
            * Examples:
            >>> Find only G-band data, which covers the wavelength range 410nm—550nm:
                BANDNAME=g
            >>> Find only G-band and I-band data, which cover the wavelength ranges 410nm—550nm 
               and 700nm—900nm respectively:
               BANDNAME=g,i

        MAG: [1 or 2-value array]
            The MAG parameter specifies a range in which the magnitude of ZTF objects of interest 
            must lie.
            // Implemented as a filter on the "medianmag" column of the ZTF objects table.
            * Examples:
            >>> MAG=17.0 17.7
            
        NUM_OBS: [int]
            The NUM_OBS parameter specifies the minimum number of observation epochs required
            of any ZTF object of interest.
            // Implemented as a filter on the "nobs" column of the ZTF objects table.
            * Examples:
            >>> NUM_OBS=5

        TIME: [1 or 2-value array]
            The TIME parameter specifies the date-time range for which lightcurve data is 
            to be retrieved. The range is unlimited by default. 
            Range endpoint(s) are interpreted as Modified Julian Dates (MJD).
            // Implemented as a filter on the "mjd" field in the ZTF lightcurve collection.
            * Examples:
            >>> Retrieve only data in the MJD range 55555.5—555678.9:
                TIME=55555.5 55678.9
            >>> Retrieve only data from at or before the MJD time 55555.5:
                TIME=-Inf 55555.5
            >>> Retrieve only data from the MJD instant 55555.5:
                TIME=55555.5

        BAD_CATFLAGS_MASK: [int]
            The BAD_CATFLAGS_MASK parameter specifies a bitmask used to exclude lightcurve points 
            with at least one of the indicated catflag bits set. 
            (See e.g. Section 10.3 of The ZTF Science Data System Explanatory Supplement 
            for a description of these bits.)
            // Implemented as a filter on the "catflags" field in the ZTF lightcurve collection.
            * Examples:
            >>> Exclude any lightcurve point whose catflag value indicates at least one of the 
                data issues associated to bits 0-3.
                BAD_CATFLAGS_MASK=15
                
        COLLECTION: [string]
            The COLLECTION parameter identifies the set of ZTF lightcurve files from which data 
            will be returned, as well as the associated ZTF objects table. 
            The default collection corresponds to the most recent public release. 
            Currently supported values are "ztf" (login required) and "ztf_dr1".
            * Examples:
            >>> COLLECTION=ztf_dr1

        FORMAT: [string]
            // Currently only CVS available with this method.
            The FORMAT parameter indicates the desired format of the output table. 
            Possible values are VOTABLE, IPAC_TABLE, HTML, CSV (the default), and TSV 
            (case-insensitive).
            * Examples:
            >>> FORMAT=VOTABLE
            
        Returns
        -------
        LCQuery
        """
        this = LCQuery()
        if cookies is None:
            from .io import get_cookie
            if auth is None:
                from .io import _load_id_
                auth = _load_id_("irsa")
            cookies = get_cookie(*auth)

        input_query = {k.upper():v for k,v in kwargs.items()}
        
        # - Build query
        if "FORMAT" in input_query.keys():
            warnings.warn("Only csv format implemented. Input 'FORMAT' ignored")
            _ = input_query.pop("FORMAT")

        this.build_query(**{**{"FORMAT":"CSV"},**input_query})
        
        this.data = read_csv( StringIO(
            requests.get( this.query_url, cookies=cookies).content.decode('utf-8')
            ) )
        
        return this
    
            
    def build_query(self, **kwargs):
        """ """
        self.query_url  = build_url( **kwargs )
        self.query_prop = kwargs


    def show_lc(self, showtoday=False, show_upperlimits=False):
        """ """
        from astropy.time import Time
        import matplotlib.dates as mdates
        import matplotlib.pyplot as mpl
        
        # ----------- #
        # Global      #
        # ----------- #
        lc_dataframe = self.data
        prop = dict(marker="o", mec="0.7", ms=8, ecolor="0.7", ls="None", zorder=4)

        # ----------- #
        #
        # ----------- #
        fig = mpl.figure(figsize=[6,4])
        ax = fig.add_subplot(111)
        for filter_ in np.unique(lc_dataframe["filtercode"]):
            d = lc_dataframe[lc_dataframe["filtercode"]==filter_]
            if filter_ in FILTER_CODE:
                prop["color"] = np.asarray(FILTER_COLORS)[np.where(np.asarray(FILTER_CODE)==filter_)][0]
            else:
                prop["color"] = "0.7"
         
            ax.errorbar(d["mjd"], d["mag"], 
                                yerr=d["magerr"],
                                **prop)
        
        ax.invert_yaxis()
        if show_upperlimits:
            for filter_ in np.unique(lc_dataframe["filtercode"]):
                d = lc_dataframe[lc_dataframe["filtercode"]==filter_]
                if filter_ in FILTER_CODE:
                    color = np.asarray(FILTER_COLORS)[np.where(np.asarray(FILTER_CODE)==filter_)][0]
                else:
                    color = "0.7"
         
                ax.errorbar(d["mjd"], d["limitmag"], 
                            yerr=0.1, lolims=True,marker="None", ls="None", 
                            color=color, alpha=0.1,
                            )

        if showtoday:
            today_color = "0.7"
            import datetime
            from astropy import time
            today = time.Time(datetime.date.today().isoformat(),format="iso").mjd
            ax.axvline(today, ls="--", color=today_color, zorder=1, lw=1)
            ax.text(today, ax.get_ylim()[0]-0.05, "Today", va="bottom", ha="right", rotation=90, color=today_color)
        
        # - Labels
        ax.set_xlabel("MJD")
        ax.set_ylabel("mag")
    # ================= #
    #  Properies        #
    # ================= #



#! /usr/bin/env python
#


""" Library to convert 'meta' information into URL assuming: 
https://irsa.ipac.caltech.edu/docs/program_interface/ztf_metadata.html
"""
import os
import warnings
import numpy as np
# ========================== #
#                            #
#   Generic Query            #
#                            #
# ========================== #

KNOWN_SCIENCE_SUFFIXES = {
                      "sciimg.fits":"(primary science image)",
                      "mskimg.fits":"(bit-mask image)",
                      "psfcat.fits":"(PSF-fit photometry catalog)",
                      "sexcat.fits":"(nested-aperture photometry catalog)",
                      "sciimgdao.psf":"(spatially varying PSF estimate in DAOPhot's lookup table format)",
                      "sciimgdaopsfcent.fits":"(PSF estimate at science image center as a FITS image)",
                      "sciimlog.txt":"(log output from instrumental calibration pipeline)",
                      "scimrefdiffimg.fits.fz":"(difference image: science minus reference; fpack-compressed)",
                      "diffimgpsf.fits":"(PSF estimate for difference image as a FITS image)",
                      "diffimlog.txt":"(log output from image subtraction and extraction pipeline)",
                      "log.txt":"(overall system summary log from realtime pipeline)"}

# SOURCES 
DATA_BASEURL   = "https://irsa.ipac.caltech.edu/ibe/data/ztf/products/"
ZTFIRSA_BASE = ["/ref/","/sci/","/raw/","/cal/"]


from .io import LOCALSOURCE, CCIN2P3_SOURCE
DEFAULT_SOURCE = DATA_BASEURL

FILTERS = {"zg":1,"zr":2,"zi":3, "OO":None}


# ================== #
#  Filename parsing  #
# ================== #
def parse_filename(filename):
    """ """
    kind = filename_to_kind(filename)
    if kind == "cal":
        return parse_calfilename(filename)
    if kind == "raw":
        return parse_rawfilename(filename)
    if kind == "sci":
        return parse_scifilename(filename)

    raise ValueError(f"Cannot parse the file {filename}")

def filename_to_kind(filename):
    """ """
    filesplit = os.path.basename(filename).split("_")
    if len(filesplit) == 6:
        if "_hifreq" in filename or "_bias" in filename:
            return "cal"
        else:
            return "raw"
        
    elif len(filesplit)>=8:
        return "sci"

    raise ValueError(f"Cannot parse the file {filename} ; remark 'ref' not implemented.")

def parse_calfilename(filename):
    """ """
    _, fileday, filtercode, ccd_, qid_, suffix = os.path.basename(filename).split("_")
    year,month, day = fileday[:4],fileday[4:6], fileday[6:]
    ccdid = int(ccd_.replace("c",""))
    qid = int(qid_.replace("q",""))
    return {"year":year,
            "month":month,
            "day":day,
            "fileday":fileday,
            "ccdid":ccdid,
            "qid":qid,
            "rcid":ccdid_qid_to_rcid(ccdid,qid),
            "filtercode":filtercode,
            "filterid":FILTERS[filtercode],
            "kind":"cal",
            "type":suffix.split(".")[0]}

    
def parse_rawfilename(filename):
    """ """
    _, filefracday, paddedfield, filtercode, ccd_, suffix_ = os.path.basename(filename).split("_")
    year,month, day, fracday = filefrac_to_year_monthday_fracday(filefracday)
    ccdid = int(ccd_.replace("c",""))
    return {"year":year,
            "month":month,
            "day":day,
            "filefracday":filefracday,
            "paddedfield":paddedfield,
            "field":int(paddedfield),
            "ccdid":ccdid,
            "filtercode":filtercode,
            "filterid":FILTERS[filtercode],
            "kind":"raw",
            "suffix":suffix_}

    
def parse_scifilename(filename):
    """ """
    from .fields import ccdid_qid_to_rcid
    
    _, filefracday, paddedfield, filtercode, ccd_, imgtypecode, qid_, *suffix_ = os.path.basename(filename).split("_")
    year,month, day, fracday = filefrac_to_year_monthday_fracday(filefracday)

    ccdid = int(ccd_.replace("c",""))
    qid = int(qid_.replace("q",""))
    return {"year":year,
            "month":month,
            "day":day,
            "filefracday":filefracday,
            "paddedfield":paddedfield,
            "field":int(paddedfield),
            "ccdid":ccdid,
            "qid":qid,
            "rcid":ccdid_qid_to_rcid(ccdid,qid),
            "filtercode":filtercode,
            "filterid":FILTERS[filtercode],
            "kind":"sci",
            "suffix":"_".join(suffix_)}

# ================== #
#  Building the URL  #
# ================== #
def _localsource_to_source_(localfilepath, source=None):
    """ """
    filesource = _source_to_location_(source)
    if LOCALSOURCE not in localfilepath:
        raise ValueError("Cannot parse the location of the given file. %s, %s expected to be part of the path"%(localfilepath,LOCALSOURCE))

    if np.any([k in localfilepath for k in ZTFIRSA_BASE]):
        return localfilepath.replace(LOCALSOURCE, filesource), "irsa"
    else:
        warnings.warn("Cannot recover the online url, only IRSA ZTF files are implemented.")
        return None, "None"
    
def _source_to_location_(source):
    """ convert flexible source naming into actual source path """
    if source is None: source = "irsa"
    source = source.lower()

    if source in ["none"]:
        return ""
    
    if source in [DATA_BASEURL, LOCALSOURCE] or "/" in source:
        return source
    
    if source in ['irsa', 'ipac', "web", "online", "ztf"]:
        return DEFAULT_SOURCE
        
    if source in ["local", "cmp", "computer"]:
        return LOCALSOURCE
    
    if source in ["cc", "ccin2p3"]:
        return CCIN2P3_SOURCE
    
    if source in ['nersc']:
        raise ValueError("NERCS sourcing not ready yet.")
    
    raise ValueError("Cannot parse the given source %s"%source)


# ------------- #
#  Sci & Raw    #
# ------------- #
def science_path(year, month, day, fracday,
                paddedfield,
                filtercode, paddedccdid, qid,
                imgtypecode="o", suffix="sciimg.fits",
                source="", verbose=False, check_suffix=True):
    """ 
    suffix: [string]
        What kind of data do you want?
        - sciimg.fits (primary science image) [used if suffix is None]
        - mskimg.fits (bit-mask image)
        - psfcat.fits (PSF-fit photometry catalog)
        - sexcat.fits (nested-aperture photometry catalog)
        - sciimgdao.psf (spatially varying PSF estimate in DAOPhot's lookup table format)
        - sciimgdaopsfcent.fits (PSF estimate at science image center as a FITS image)
        - sciimlog.txt (log output from instrumental calibration pipeline)
        - scimrefdiffimg.fits.fz (difference image: science minus reference; fpack-compressed)
        - diffimgpsf.fits (PSF estimate for difference image as a FITS image)
        - diffimlog.txt (log output from image subtraction and extraction pipeline)
        - log.txt (overall system summary log from realtime pipeline)
        
    filtercode: [2 digit string]
        The filter name ('zg','zr','zi', or 'OO' for Bias)
        
    paddedccdid: [2 digit string]
        The ccd id [01,02,03....15,16]

    qid: [1 digit string]
        Which quadran [1,2,3,4]
    
    """
    if verbose:
        print(f"science_path: {locals()}")
    source = _source_to_location_(source)
    if suffix is None:
        suffix="sciimg.fits"
    elif suffix not in KNOWN_SCIENCE_SUFFIXES.keys() and check_suffix:
        raise ValueError(f"Unkwown suffix {suffix} for 'sci' exposures", "\n known suffixes: \n", KNOWN_SCIENCE_SUFFIXES)
    
    filefracday = "".join([year+month+day+fracday])
    file_ = 'sci/'+year+'/'+month+day+'/'+fracday+'/ztf_'+filefracday+'_'+paddedfield+'_'+filtercode+'_c'+paddedccdid+'_'+imgtypecode+'_q'+qid+'_'+suffix
    return os.path.join(source,file_)
    
def raw_path(year, month, day, fracday,
            paddedfield,
            filtercode, paddedccdid, imgtypecode="o",
            source=""):
    """ 

    Parameters
    ----------
    filtercode: [2 digit string]
        The filter name ('zg','zr','zi', or 'OO' for Bias)
        
    paddedccdid: [2 digit string]
        The ccd id [01,02,03....15,16]

    imgtypecode: [char]
        Currently, the possible values for "imgtypecode" pertaining to raw image data files are:
        - o = science (on-sky); see example above.
        - b = bias calibration image
        - d = dark calibration image
        - f = dome/screen flatfield calibration image

    """
    source = _source_to_location_(source)
    filefracday = "".join([year+month+day+fracday])
    file_ = 'raw/'+year+'/'+month+day+'/'+fracday+'/ztf_'+filefracday+'_'+paddedfield+'_'+filtercode+'_c'+paddedccdid+'_'+imgtypecode+'.fits.fz'
    return os.path.join(source,file_)

# ------------- #
#  Calibration  #
# ------------- #
def calibration_path(caltype,
                    year, month, day,
                    filtercode,
                    paddedccdid, qid, suffix=None,
                    source="", check_suffix=True):
    """ 
    Parameters:
    -----------
    caltype: [string]
        Should be 'bias' or 'hifreqflat'
    
    filtercode: [2 digit string]
        The filter name ('zg','zr','zi', or 'OO' for Bias)
        
    paddedccdid: [2 digit string]
        The ccd id [01,02,03....15,16]

    qid: [1 digit string]
        Caltype quadran [1,2,3,4]

    suffix: [string] -optional-
        What kind of calibration file you want for the given `caltype`
        - None: returns `caltype`.fits
        - log:  returns `caltype`log.txt
        - unc:  returns `caltype`unc.fits
    """
    # ================ #
    #  INPUT TESTING   #
    # ================ #
    source = _source_to_location_(source)
    caltype = caltype.lower() # case insensitive
    if caltype.lower() not in ["bias", "hifreqflat"] and check_suffix:
        raise ValueError(f"Unknown `caltype` {caltype}, caltype must be  'bias' or 'hifreqflat'")

    if suffix is None or suffix in ["basic",".fits","fits"]:
        suffix = caltype+".fits"
    elif suffix in ["log", "log.txt"]:
        suffix = caltype+"log.txt"
    elif suffix in ["unc", "uncorrected","unc.fits"]:
        suffix = caltype+"unc.fits"
    else:
        raise ValueError("Unknow suffix %s must be: basic/None or log or unc")
    
    # ================ #
    #  Build URL       #
    # ================ #
    filestartdate = "".join([year,month,day])
    file_ = 'cal/'+year+'/'+month+day+'/%s/'%caltype.lower()+filtercode+'/ccd'+paddedccdid+'/q'+qid+'/ztf_'+filestartdate+'_'+filtercode+'_c'+paddedccdid+'_q'+qid+'_%s'%suffix
    return os.path.join(source,file_)

# ------------- #
#  References   #
# ------------- #
def reference_path(paddedfield,
                  filtercode,
                  paddedccdid, qid,
                  fieldprefix="000",
                  suffix=None,
                  source="", verbose=True):
    """   

    filtercode: [2 digit string]
        The filter name ('zg','zr','zi', or 'OO' for Bias)
        
    paddedccdid: [2 digit string]
        The ccd id [01,02,03....15,16]

    qid: [1 digit string]
        Which quadran [1,2,3,4]

    suffix: [string]
        Could be:
        - log.txt
        - refcov.fits
        - refimg.fits
        - refimlog.txt
        - refpsfcat.fits
        - refsexcat.fits
        - refunc.fits

    """
    if verbose:
        print(f"reference_path: {locals()}")

    if suffix is None:
        suffix = "refimg.fits"
    source = _source_to_location_(source)
    file_ = 'ref/'+fieldprefix+'/field'+paddedfield+'/'+filtercode+'/ccd'+paddedccdid+'/q'+qid+'/ztf_'+paddedfield+'_'+filtercode+'_c'+paddedccdid+'_q'+qid+'_%s'%suffix
    return os.path.join(source,file_)

# ============= #
#   TOOLS       #
# ============= #
def build_filename_from_dataframe(dataframe, suffix="sciimg.fits"):
    """ The dataframe must contains:
    filefracday, fieldid, ccdid, qid and filterid

    Returns
    ------
    pandas.Series (name of the files (not fullpath))
    """
    filefrac    = dataframe["filefracday"].astype("int").astype("str")
    paddedfield = dataframe["fieldid"].astype("int").astype("str").str.pad(6, fillchar="0")
    paddedccd = "c"+dataframe["ccdid"].astype("int").astype("str").str.pad(2, fillchar="0")
    paddedqid = "q"+dataframe["qid"].astype("int").astype("str")
    filtername = dataframe["filterid"].apply(lambda x: "zg" if x==1 else "zr" if x==2 else "zi")
    return "ztf"+"_"+filefrac+"_"+paddedfield+"_"+filtername+"_"+paddedccd+"_"+"o"+"_"+paddedqid+"_"+suffix


def filename_to_url(filename, suffix=None, source="irsa", verbose=False, check_suffix=True):
    """ Generic high level function that first detects the inputfile kind and then 
    calls the associated function ; e.g., filename_to_scienceurl or filename_to_calurl.

    NB:
    suffix -> imgtypecode for raw images.
    """
    kind = filename_to_kind(filename)
    if kind == "sci":
        return filename_to_scienceurl(filename, suffix=suffix, source=source, verbose=verbose, check_suffix=check_suffix)
    
    if kind == "raw":
        return filename_to_rawurl(filename, imgtypecode=suffix, source=source, verbose=verbose, check_suffix=check_suffix)
    
    if kind == "cal":
        return filename_to_calurl(filename, suffix=suffix, source=source, verbose=verbose, check_suffix=check_suffix)

    raise ValueError(f"Cannot parse the file {filename}")
    
def filename_to_scienceurl(filename, suffix=None, source="irsa", verbose=False, check_suffix=True):
    """ 
    """
    _, filefracday, paddedfield, filtercode, ccd_, imgtypecode, qid_, *suffix_ = os.path.basename(filename).split("_")
    suffix_ = "_".join(suffix_)
    
    year,month, day, fracday = filefrac_to_year_monthday_fracday(filefracday)
    paddedccdid = ccd_.replace("c","")
    qid = qid_.replace("q","")
    
    if suffix is None:
        suffix = suffix_
        
    return science_path(year, month, day, fracday,
                        paddedfield, filtercode,
                        paddedccdid, qid, # added in input
                        imgtypecode=imgtypecode, suffix=suffix,
                        source=source, verbose=verbose,
                        check_suffix=check_suffix)
    
def filename_to_calurl(filename, suffix=None, source="irsa", verbose=False, check_suffix=True):
    """ 
    """
    _, date, filtercode, ccd_, qid_, *suffix_ = os.path.basename(filename).split("_")
    year,month,day = date[:4],date[4:6],date[6:]
    suffix_ = "_".join(suffix_)
    caltype = suffix_.split(".")[0]
    paddedccdid = ccd_[1:]
    qid = qid_[1:]
    if suffix is None:
        suffix = suffix_
    
    return calibration_path(caltype,
                                year, month, day,
                                filtercode,
                                paddedccdid, qid, suffix=suffix,
                                source=source, check_suffix=True)


def filename_to_rawurl(filename, imgtypecode=None, source="irsa", verbose=False, check_suffix=True):
    """ 
    """
    _, filefracday, paddedfield ,filtercode, ccd_, suffix_ =  os.path.basename(filename).split("_")
    year, month, day, fracday = filefrac_to_year_monthday_fracday(filefracday)
    imgtypecode_ = suffix_.split(".")[0]
    paddedccdid = ccd_[1:]
    if imgtypecode is None:
        imgtypecode = imgtypecode_
        
    return raw_path(year, month, day, fracday,
                             paddedfield, filtercode, paddedccdid, imgtypecode=imgtypecode,
                             source=source)


def filename_to_refurl(filename, suffix, source="irsa", verbose=False):
    """ 
    suffix: [string]
        Could be:
        - log.txt
        - refcov.fits
        - refimg.fits
        - refimlog.txt
        - refpsfcat.fits
        - refsexcat.fits
        - refunc.fits
    """
    _, filefracday, paddedfield, filtercode, ccd_, imgtypecode, qid_, suffix_ = os.path.basename(filename).split("_")
    year,month, day, fracday = filefrac_to_year_monthday_fracday(filefracday)
    paddedccdid = ccd_.replace("c","")
    qid = qid_.replace("q","")
    
    return reference_path(paddedfield,
                                   filtercode,
                                   paddedccdid, qid,
                                   fieldprefix=paddedfield[:3],
                                   suffix=suffix,
                                   source=source)


def filefrac_to_year_monthday_fracday(filefracday):
    """ split the name as YYYY, MM,DD, FFFFFF
    FFFFFF (fraction in the day)"""
    return filefracday[:4], filefracday[4:6], filefracday[6:8],filefracday[8:]


def fileroot_to_science_url(fileroot, paddedccdid, qid,
                            imgtypecode="o", suffix="sciimg.fits", source="",
                            verbose=False, check_suffix=True):
    """ 
    Parameters
    ----------

    """
    _, filefracday, paddedfield, filtercode = fileroot.split("_")
    year,month, day, fracday = filefrac_to_year_monthday_fracday(filefracday)
    return science_path(year, month, day, fracday,
                paddedfield, filtercode,
                paddedccdid, qid, # added in input
                imgtypecode=imgtypecode, suffix=suffix,
                source=source, verbose=verbose, check_suffix=check_suffix)


def fileroot_to_raw_url(fileroot, paddedccdid,
                        imgtypecode="o", source="", verbose=False):
    """ 
    Parameters
    ----------

    """
    _, filefracday, paddedfield, filtercode = fileroot.split("_")
    year, month, day, fracday = filefrac_to_year_monthday_fracday(filefracday)

    return raw_path(year, month, day, fracday,
            paddedfield,
            filtercode, paddedccdid, imgtypecode=imgtypecode,
            source=source)
    




""" Generic projection and plotting tools """

import numpy as np
_DEG2RA = np.pi / 180


def is_running_from_notebook():
    """ Test if currently ran in notebook """
    return running_from() == "notebook"

def running_from():
    """  Where is the code running from ?"""
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return "notebook"   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return "terminal"  # Terminal running IPython
        else:
            return "other"  # Other type (?)
    except NameError:
        return None 
    
def avoid_duplicate(string_array):
    """ """
    string_array = np.asarray(string_array, dtype=object)
    occurence = {k:v for k,v in zip(*np.unique(string_array, return_counts=True))}
    for name_,nname_ in occurence.items():
        if nname_>1:
            string_array[np.in1d(string_array,name_)] = [name_+f"{i+1}" for i in range(nname_)]
            
    return np.asarray(string_array, dtype="str")

##############################
#                            #
#  Convertion Tools          #
#                            #
##############################
# ---------------------- #
#  Sphere to Cartesian   #
# ---------------------- #
def sph2cart(vec):
    """
    Convert vector in spherical coordinates (r, theta, phi ; angles in degrees)
    to Cartesian coordinates [x,y,z].
    
    Returns
    -------
    x,y,z
    """
    v, l, b = vec[0], vec[1]*_DEG2RA, vec[2]*_DEG2RA
    return np.asarray([v*np.cos(b)*np.cos(l), 
                       v*np.cos(b)*np.sin(l), 
                       v*np.sin(b)])    

def cart2sph(vec):
    """
    Convert vector in Cartesian coordinates [x,y,z] to spherical coordinates [r, theta, phi]
    (angles in degrees).
    """
    x, y ,z = vec
    v = np.sqrt(x**2 + y**2 + z**2)
    return np.array([v,
                    (np.arctan2(y,x) / _DEG2RA + 180) % 360 - 180, 
                     np.arcsin(z/v) / _DEG2RA])
     
# ---------------------- #
#  Rotation              #
# ---------------------- #

def rot_xz(v, theta):
    """
    Rotate Cartesian vector v [x,y,z] by angle theta around axis (0,1,0)
    """
    return np.asarray([v[0]*np.cos(theta*_DEG2RA) - v[2]*np.sin(theta*_DEG2RA),
                      v[1],
                      v[2]*np.cos(theta*_DEG2RA) + v[0]*np.sin(theta*_DEG2RA)])

def rot_xz_sph(l, b, theta):
    """
    Rotate Spherical coordinate (l,b = theta, phi) by angle theta around axis (0,1,0)
    """
    v_rot = rot_xz(sph2cart([1,l,b]), theta)
    return cart2sph(v_rot)[1:]





#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
 ZTF filter informations.
"""
import os
import numpy as np
import pandas

_FILTER_LOC = os.path.dirname(os.path.realpath(__file__))+"/data/"

P48_FILTER = {"ztf:g":{"band":"g", "telescope":"p48", "instrument":"ztf"},
              "ztf:r":{"band":"r", "telescope":"p48", "instrument":"ztf"},
              "ztf:i":{"band":"i", "telescope":"p48", "instrument":"ztf"},
                }
try:
    import sncosmo
    _HAS_SNCOSMO = True
except:
    _HAS_SNCOSMO = False
    

def get_p48_filter(band):
    """ returns a pandas dataframe for the corresponding filter 
    Parameters
    ----------
    band: [str]
        P48 filter name g, r or i
        
    Returns
    -------
    pandas DataFrame (Wavelength, Effective Transmission)
    """
    
    band_ = band.lower()
    if band_ in P48_FILTER.keys():
        band_ = band_.replace("ztf:","")
    elif band_ not in ["g","r","i"]:
        raise ValueError("Only g, r, or i filter accepted, '%s' given"%band_)
    
    return pandas.read_csv(_FILTER_LOC+"filter_p48_%s.dat"%band_, sep=" ")

        
def get_p48_bandpass(band):
    """ """
    try:
        import sncosmo
    except ImportError:
        raise ImportError("You do not have sncosmo. Please install it (pip install sncosmo)")

    band_ = band.lower()
    if band_ in ["g","r","i"]:
        band_ = "z"+band
    elif band_ not in P48_FILTER.keys():
        raise ValueError("Only (ztf:)g, (ztf:)r, or (ztf:)i  filter accepted, '%s' given"%band_)
    
    try:
        return sncosmo.get_bandpass(band_)
    except:
        raise Exception("If seems that sncosmo does not know '%s' ; have you ran load_p48_filters_to_sncosmo() ?"%band_)
        
def load_p48_filters_to_sncosmo(bands="*", basename="ztf:"):
    """ register the given bands ['g','r' or 'i'] as basename+`band`"""
    try:
        import sncosmo
    except ImportError:
        raise ImportError("You do not have sncosmo. Please install it (pip install sncosmo)")
    
    bands_ = ["g","r","i"] if bands in ["all","*"] else np.atleast_1d(bands)
    
    for bandname in bands_:
        df_ = get_p48_filter(bandname)
        band = sncosmo.Bandpass(df_["wavelength"].values, df_["eff_trans"].values, name=basename+bandname.lower())
        sncosmo.registry.register(band, force=True)
    
if _HAS_SNCOSMO:
    load_p48_filters_to_sncosmo()
    



#! /usr/bin/env python
# -*- coding: utf-8 -*-

#import numpy as np
from astropy import time, units
    
STRINGTYPEKEYS = []
        
def build_singlequery(key, value, asstring=False):
    """ """
    if type(value) == str:
        value = [l.strip()for l in value.split(",")]
    else:
        value = [str(v) for v in np.atleast_1d(value)]
        
    if asstring:
        value = ["'%s'"%v for v in value]
        
    if len(value) == 1:
        return f"{key} = {value[0]}"

    value = ",".join(value)
    return f"{key} IN ({value})"

def _isfloat_(value):
    """ """
    try:
        _ = float(value)
        return True
    except:
        return False
    
def _get_time_(default, input_, refdate=None, format=None):
    """ """
    if default is None and input_ is None:
        return None
    
    if _isfloat_(input_):
        if refdate is None:
            raise ValueError("Cannot set relative times to undefined refdate time.")        
        return time.Time(refdate, format=format)+ float(input_)*units.day

    return  time.Time(default if input_ is None else input_, format=format) 
        
def build_query(inputargs):
    """ Builds the SQL Query based on the input args """
    #
    # - Dates
    starting_date = None
    ending_date   = None
    
    if inputargs.target is not None:
        if inputargs.targetsource.lower() in ["marshal"]:
            from ztfquery import marshal
            m = marshal.MarshalAccess.load_local()
            if inputargs.target not in m.target_sources["name"].values:
                raise ValueError(f"unknown target {inputargs.target}")
            # Overwriting
            inputargs.radec  = list(m.get_target_coordinates(inputargs.target).values[0])
            tcreation, tlast = m.get_target_jdrange(inputargs.target, format="time")
            starting_date    =  _get_time_(tcreation-30*units.day, inputargs.startdate,
                                            refdate=tcreation, format=inputargs.dateformat)
            ending_date      =  _get_time_(tlast+30*units.day, inputargs.enddate, refdate=tlast,
                                            format=inputargs.dateformat)
        else:
            raise NotImplementedError("Only targetsource marshal implemented")

    #{'kind': 'sci', 'radec': [152.7069519, 54.214881999999996], 'size': 0.01, 'caltype': None, 'sql_query': 'rcid = 9 and obsjd BETWEEN 2458537.755162037 AND 2458675.108217593'}
    if inputargs.obsdate is not None:
        inputargs.startdate = inputargs.obsdate
        inputargs.enddate = "+1"
        
    if starting_date is None:
        starting_date = _get_time_(None, inputargs.startdate, refdate=None,
                                    format=inputargs.dateformat)
    if ending_date is None:
        ending_date   = _get_time_(None, inputargs.enddate, refdate=inputargs.startdate,
                                    format=inputargs.dateformat)
        
    # - Query
    if starting_date is not None and ending_date is not None:
        timequery = f"obsjd BETWEEN {starting_date.jd} AND {ending_date.jd}"
    elif starting_date is not None:
        timequery = f"obsjd >= {starting_date.jd}"
    elif ending_date is not None:
        timequery = f"obsjd <= {ending_date.jd}"
    else:
        timequery = None
                
    # - Queries
    queries = [build_singlequery(k, getattr(inputargs,k), k in STRINGTYPEKEYS)
                for k in ["ccdid","qid","rcid","fid","field"] if getattr(inputargs,k) is not None]
    if timequery is not None:
        queries += [timequery]
    if inputargs.sqlquery is not None:
        queries += [inputargs.sqlquery]
        
    if len(queries)>=1:
        sql_query = " and ".join(queries)
    else:
        sql_query=None
        
    return sql_query

#################################
#
#   MAIN 
#
#################################
if  __name__ == "__main__":
    
    
    import argparse
    import sys
    import ztfquery
    from ztfquery import query
    # ================= #
    #   Options         #
    # ================= #
    parser = argparse.ArgumentParser(
        description="""download ZTF from IRSA using ztfquery.""",
        formatter_class=argparse.RawTextHelpFormatter)


    # // Base query
    parser.add_argument('--fromname', type=str, default=None,
                        help="Download the name assocated to the file. If you want several, just split them with a comma. Remark that you can then specify the suffix")
    
    parser.add_argument('--frommetafile', type=str, default=None,
                        help="Provide a local metatable csv file.")

    parser.add_argument('--target', type=str, default=None,
                        help="Provide a target name and this will look its ra, dec and dates to ")

    parser.add_argument('--targetsource', type=str, default="marshal",
                        help="Where should the target information come from ? ")

    # // Generic
    parser.add_argument('--suffix', type=str, default=None,
                        help="Which data should be downloaded. If you want several, just split them with a comma: e.g. sciimg.fits,psfcat.fits")
        
    parser.add_argument('--auth',  nargs=2, type=str, default=None,
                        help="Your IRSA login and password")

    parser.add_argument('--overwrite',  action="store_true", default=False,
                        help="Re-download and overwrite existing files ?")

    parser.add_argument('--verbose',  action="store_true", default=False,
                        help="Switch on the verbose mode")
    
    parser.add_argument('--nodl',  action="store_true", default=False,
                        help="Print the downloading instead of actually run it")

    
    # // Data Kind
    parser.add_argument('--kind',  type=str, default="sci",
                        help="Kind of data. Could be: sci, raw, ref or cal")

    parser.add_argument('--caltype',  type=str, default=None,
                        help="Type of Calibration files you are looking for 'bias' or 'hifreqflat'.\n *Ignored* if not '--kind raw' ")

    # // Target Oriented
    parser.add_argument('--radec', nargs=2, type=float, default=None,
                        help="RA Dec Coordinates in degree if any. *Ignored* if --target given")

    parser.add_argument('--size', type=float, default=0.01,
                        help="Size [in degree] of the cone search..\n *Ignored* if --radec or --target is not used.")

    # // Detailed Query
    parser.add_argument('--obsdate', type=str, default=None,
                        help="Date in YYYY-MM-DD format")
    
    parser.add_argument('--startdate', type=str, default=None,
                        help="Starting date. Should be understood by astropy.time.Time() ; see --dateformat \n If --target given this could also be relative [in days] like -2 for 2 days before first detection.")
    
    parser.add_argument('--enddate', type=str, default=None,
                        help="Ending date. Should be understood by astropy.time.Time() ; see --dateformat"+"\n Could also be relative [in days] like +2 for 2 days after startdate")
    
    parser.add_argument('--dateformat', type=str, default=None,
                        help="Format of of the given startdate and stopdate if any. astropy.time.Time format.")

    # // CCD, Quadran and Filters
    parser.add_argument('--ccdid', type=str, default=None,
                        help="ccdid you want. If you want several, just split them with a comma: e.g. 2,6,14")

    parser.add_argument('--qid', type=str, default=None,
                        help="quadran id you want (1->4). If you want several, just split them with a comma: e.g. 1,4")

    parser.add_argument('--rcid', type=str, default=None,
                        help="ccd-quadran combination you want (rcid = 4*(ccdid - 1) + qid - 1). If you want several, just split them with a comma: e.g. 34,45,62")

    parser.add_argument('--fid', type=str, default=None,
                        help="filterid you want. (1: ztf:g ; 2: ztf:r ; 3:ztf.i). If you want several, just split them with a comma: e.g. 1,2")

    parser.add_argument('--field', type=str, default=None,
                        help="field you want. If you want several, just split them with a comma: e.g. 754,786")

    parser.add_argument('--sqlquery', type=str, default=None,
                        help="Any metadata SQL query that will be added to the one before")

    parser.add_argument('--storemeta', type=str, default=None,
                        help="Provide the filename of where you want to save the metatable. None means no save.")

    parser.add_argument('--queryonly',  action="store_true", default=False,
                        help="Only run the load_metatable query, no download. Remark, you want to use --storemeta to save this query.")
    
    # // Download options
    parser.add_argument('--dlsource', type=str, default="IRSA",
                        help="From where the data should be downloaded ?")
    
    parser.add_argument('--dlindex', type=str, default=None,
                        help="Which index of the metatable would you like to download ? None means all")
    
    parser.add_argument('--outdir', type=str, default=None,
                        help="Where should the data be downlaoded. If None this follows the IRSA/ztfquery structure (recommended) ")
    
    parser.add_argument('--progress',  action="store_true", default=False,
                        help="Show the download progress bar.")
    
    parser.add_argument('--nprocess',  type=int, default=4,
                        help="Number of parallel downloading.")

    parser.add_argument('--nowarnings', action="store_true", default=False,
                        help="Ignore the warnings.")

    
    args = parser.parse_args()

    # Matplotlib
    if args.verbose:
        print(f"based on ztfquery {ztfquery.__version__}")
    # ================= #
    #   The Scripts     #
    # ================= #
    download_prop = dict(auth=args.auth, overwrite=args.overwrite,
                         show_progress=args.progress, 
                         nprocess=args.nprocess,verbose=args.verbose)
    #
    # From Name
    #
    if args.fromname is not None:
        from ztfquery.io import download_from_filename
        if args.verbose:
            print(f"Query from file {args.fromname}")

        for filename in args.fromname.split(","):
            if args.suffix in None:
                download_from_filename(filename, **download_prop)
            else:
                [download_from_filename(filename, suffix=s, **download_prop)
                  for s in args.suffix.split(',')]
                
        sys.exit(0)
        # STOP            
    
    #
    # Loading ZTFQuery
    #

    if args.frommetafile is not None:
        if args.verbose:
            print(f"ZTFQuery build from input metafile {args.frommetafile}")

        zquery = query.ZTFQuery.from_metafile(args.frommetafile)
    else:
        sql_query = build_query(args)
        metaquery = dict(kind=args.kind, radec=args.radec, size=args.size,
                             caltype=args.caltype, sql_query=sql_query)
        if args.verbose:
            print(metaquery)
        zquery = query.ZTFQuery.from_metaquery(**metaquery)
        if args.verbose:
            print(f"{len(zquery.metatable)} entries")

        if args.storemeta is not None:
            zquery.metatable.to_csv(args.storemeta)

    if args.queryonly:
        sys.exit(0)
    #
    # Lunch Downloading
    #
    download_prop = {**download_prop,
                     **dict(source=args.dlsource,
                            indexes=args.dlindex, download_dir=args.outdir,
                            nodl=args.nodl,
                            nprocess=args.nprocess,
                            filecheck=True,
                            erasebad=True,
                            redownload=True,
                            ignore_warnings=args.nowarnings)
                    }

    if args.verbose:
        print(f"Download starting...")
            
    if args.suffix is None:
        _ = zquery.download_data(suffix=None, **download_prop)
    else:
        _ = {s:zquery.download_data(suffix=s, **download_prop) for s in args.suffix.split(",")}
        
    if args.nodl:
        print(_)



#!/usr/bin/env python
#

import os, hashlib
import sys
import time
import pandas
import requests
import warnings
import numpy as np
LOGIN_URL = "https://irsa.ipac.caltech.edu/account/signon/login.do"

import base64

from configparser import ConfigParser
from astropy.io import fits
        
from .utils.tools import is_running_from_notebook


_SOURCEDIR = os.path.dirname(os.path.realpath(__file__))


_ENCRYPT_FILE = os.path.expanduser("~")+"/.ztfquery"
_ENCRYPTING_FILE = os.path.expanduser("~")+"/.queryirsa"


LOCALSOURCE   = os.getenv('ZTFDATA',"./Data/")
CCIN2P3_SOURCE = "/sps/ztf/data/"


# ================= #
#  High level tools #
# ================= #
def get_file(filename, suffix=None, downloadit=True, verbose=False, check_suffix=True,
                 dlfrom="irsa", overwrite=False, maxnprocess=4, exist=True, test_file=True,
                 squeeze=True, show_progress=True, client=None, wait=None,
                 fill_notexist="None", **kwargs):
    """ Get full path associate to the filename.
    If you don't have it on your computer, this downloads it for you.

    Parameters
    ----------
    filename: [string]
        name of the file you want.
        raw, cal and sci filenames are accepted. 
        
    suffix: [string] -optional-
        actual suffix of the file you want. By default it is that of the filename
        but you can request associated files with other suffix.
        For instance:
        if filename = ztf_20190917468333_000698_zi_c03_o_q2_sciimg.fits and suffix='mskimg.fits'
        you will be looking for ztf_20190917468333_000698_zi_c03_o_q2_mskimg.fits.
        - If you input a raw file, then suffix is used as 'imgtypecode'

    downloadit: [bool] -optional-
        If you do not have the file locally, shall this download it for you ?
        
    wait: [None/string/float] -optional-
        Waiting time for dwnloaded the file. It helps mostly when dask is massively multi-downloading.
        - None: if Client, wait -> len(to_be_downloaded)/100 (so 100 per second) ; corresponds to wait="100"
        - "None": no limit
        - "float_in_string": waiting time reported to the size of file to download: len(to_be_downloaded)/wait
        - float: absolute time to wait.

    **kwargs goes to download_from_filename

    Returns
    -------
    fullpath (or None if not data)
        
    """
    from .buildurl import filename_to_url
    local_filenames = np.asarray([filename_to_url(filename_, suffix=suffix_,
                                                    verbose=verbose,
                                                    source="local", check_suffix=check_suffix)
                   for filename_ in np.atleast_1d(filename)
                   for suffix_ in np.atleast_1d(suffix)])
    if not exist:
        return local_filenames
    
    # local_filenames
    if overwrite:
        flag_todl = np.asarray(np.ones(len(local_filenames)), dtype="bool")
    else:
        flag_todl = np.asarray([(not os.path.isfile(f_)) or
                                    (test_file and ".fits" in f_ and _is_fitsfile_bad_(f_))
                                for f_ in local_filenames])

    # DL if needed (and wanted)
    if np.any(flag_todl) and downloadit:
        if client is not None and wait is None:
            wait = "100"
        if type(wait) is str:
            if wait == "None":
                wait = None
            else:
                try:
                    wait = len(local_filenames[flag_todl])/float(wait)
                except:
                    warnings.warn(f"cannot parse the inpout waiting time {wait} -> None used.")
                    wait = None            
        
        f_ = download_from_filename(local_filenames[flag_todl], show_progress=show_progress,
                                    host=dlfrom, overwrite=True, maxnprocess=maxnprocess,
                                    client=client, wait=wait,
                                    check_suffix=check_suffix)
        if client is not None:
            from dask.distributed import wait as dwait
            _ = dwait(f_)
            
    # - Output
    if fill_notexist != "None":
        local_filenames = [f if os.path.isfile(f) else fill_notexist
                               for f in local_filenames]
            
    if len(local_filenames)==1 and squeeze:
        return local_filenames[0]
    
    return local_filenames


def filefracday_to_local_rawdata(filefracday, ccdid="*"):
    """ """
    
    from glob import glob
    from .buildurl import filefrac_to_year_monthday_fracday
    filefracday = str(filefracday)
    year, month, day, fracday = filefrac_to_year_monthday_fracday(filefracday)
    if ccdid in ["*","all"]:
        cstring = "*"
    else:
        cstring = f"_c{ccdid:02d}_"
    return np.sort( glob(os.path.join( LOCALSOURCE, "raw",year, f"{month}{day}",fracday,f"ztf_{filefracday}*{cstring}*")) )

def bulk_get_file(filenames, client=None, suffix=None, as_dask="delayed", **kwargs):
    """ 
    Parameters
    ----------
    as_dask: [string] -optional-
         could be
         - delayed
         - futures
         - gathered
    """
    import dask
    d_files = [dask.delayed( get_file )(filename, suffix=suffix, show_progress=False,
                                            maxnprocess=1, **kwargs)
                   for filename in filenames]
    if as_dask == "delayed":
        return d_files

    futures = client.compute(d_files)
    if as_dask == "futures":
        return futures
    if as_dask in ["gather", "gathered"]:
        return client.gather(futures)
    raise ValueError(f"Cannot parse the given as_dask {as_dask}")
    

def get_filedataframe(filenames):
    """ get a dataframe of the files"""
    import pandas
    fileserie = pandas.Series(filenames, name="filename")
    
    fdata     = pandas.DataFrame.from_records( fileserie.apply( parse_filename ))
    fdata["isfile"] = fileserie.apply( os.path.isfile)
    merged = fdata.merge(fileserie, left_index=True, right_index=True)
    return merged

def parse_filename(filename, as_serie=True):
    """ """
    from .buildurl import parse_filename
    if as_serie:
        return pandas.Series( parse_filename(filename) )
    return parse_filename(filename)

def download_from_filename(filename, suffix=None, verbose=False, overwrite=False,
                               auth=None, nodl=False, host="irsa", maxnprocess=4,
                               show_progress=True, check_suffix=True,
                               client=None, wait=None, **kwargs):
    """ Download the file associated to the given filename """
    if host not in ["irsa", "ccin2p3"]:
        raise ValueError(f"Only 'irsa' and 'ccin2p3' host implemented: {host} given")
    
    from .buildurl import filename_to_url
    if auth is None:
        auth = _load_id_(host)

    remote_filename = []
    local_filename  = []
    for file_ in np.atleast_1d(filename):
        remote_filename.append(filename_to_url(file_, suffix=suffix, verbose=verbose,
                                               source=host, check_suffix=check_suffix))
        local_filename.append(filename_to_url(file_, suffix=suffix, verbose=verbose,
                                                source="local", check_suffix=check_suffix))
        
    if nodl:
        return [remote_filename, local_filename]

    if host == "ccin2p3":
        return [CCIN2P3.scp(remote_filename_, local_filename_, auth=auth)
                    for remote_filename_, local_filename_ in zip(remote_filename,local_filename)]
        
    else:
        nprocess = np.min([maxnprocess, len(local_filename)])

        return download_url(remote_filename,
                            local_filename,
                            nprocess=nprocess,
                            client=client, wait=wait,
                            overwrite=overwrite,verbose=verbose,
                            cookies = get_cookie(*auth),
                            show_progress=show_progress, 
                            **kwargs)


def _parse_filename_(filename, builddir=False, squeeze=True, exists=False):
    """ """
    print("IS THIS USED ?")
    from glob import glob
    directory = os.path.dirname(filename)
    #basename  = os.path.basename(filename).split(".")[0]
    #extension = filename.split(".")[-1]
    
    if builddir:
        os.makedirs(directory, exist_ok=True)

    # unique object
    if "*" in filename and not exists:
        warnings.warn("apparent variable conflict, filename contains '*', but exists is False.")
        
    localfile = glob(filename) if exists else np.atleast_1d(filename)
    if squeeze:
        if len(localfile)==0:
            return None
        if len(localfile)==1:
            return localfile[0]
        
    return localfile
# ================= #
#  Crypting         #
# ================= #

def _load_id_(which, askit=True,  token_based=False):
    """ returns login information for the requested enty"""
    import base64
    config = ConfigParser()
    config.read( _ENCRYPT_FILE )
    

    if which in ["fritz"]:
        token_based = True
        
    if which not in config.sections():
        if not askit:
            raise AttributeError(f"No {which} account setup. Add then in .ztfquery or run ztfquery.io.set_account({which})")
        else:
            warnings.warn(f"No {which} account setup, please provide it")
            set_account(which, token_based=token_based)
            config = ConfigParser()
            config.read( _ENCRYPT_FILE )

    if not token_based:
        return config[which.lower()]["username"], base64.b64decode(config[which.lower()]["password"][2:-1]).decode("utf-8")
    
    return base64.b64decode(config[which.lower()]["token"][2:-1]).decode("utf-8")

    
def set_account(which, username=None, password=None, token=None, test=True, force=False, token_based=False):
    """ Setup the username and password (simply encrypted!) for the given `which` account. 
    Saved in ~/.ztfquery
    """
    import base64
    import getpass
    config = ConfigParser()
    config.read( _ENCRYPT_FILE )
    
    
    if which in ["fritz"]:
        token_based = True

    if token_based:
        if token is None:
            token = input(f"Enter your {which} token:")
    else:
        # - Name & Password
        if username is None:
            username = input(f'Enter your {which} login: ')
            
        if password is None:
            password = getpass.getpass()

    #
    # -> Starting tests
    if test:
        wrong_ = False
        if which == "irsa":
            if not test_irsa_account([username, password]):
                warnings.warn("The irsa_test for you account returns False. Most likely you provided incorrect logins")
                wrong_ = True
        else:
            if not token_based:
                warnings.warn(f"No test designed for {which}. Cannot test if logins are correct.")
            else:
                warnings.warn(f"No test designed for {which}. Cannot test if token is correct.")
            
        if wrong_ and not force:
            if not token_based:
                raise ValueError(f"Bad username/passworg for {which}. force=False so the logins are not stored. ")
            else:
                raise ValueError("Bad token for {which}. force=False so the logins are not stored.")
    # <- end of tests
    #
    if not token_based:
        password_ = base64.b64encode( password.encode("utf-8") )
        config[which.lower()] = {"username":username, "password": password_ }
    else:
        token_ = base64.b64encode( token.encode("utf-8") )
        config[which.lower()] = {"token":token_}
        
    with open( _ENCRYPT_FILE , 'w') as configfile:
        config.write(configfile)


#
# TEST
#
# - Password testing
def test_irsa_account(auth=None):
    """  returns True if the IRSA account is correctly set. """
    if auth is None:
        auth = _load_id_("irsa")
    return ".ipac.caltech.edu" in get_cookie(*auth)._cookies



# - File testing
def get_localfiles(extension="*", startpath=None):
    """ Look for all file with the given extension recursively starting from `startpath`.
    (based on glob)
    
    Parameters
    ----------
    extension: [string] -optional-
        All the 'file.{}'.format(extension) will be looked at.
        (first '.' ignored such that extension='.fits' is equivalent to extension='fits')

    startpath: [None or path] -optional-
        From which directory does this start to look at.
        If None: $ZTFDATA (stored as io.LOCALSOURCE) will be used.

    Returns
    -------
    list of file.
    """
    from glob import glob
    if startpath is None:
        startpath = LOCALSOURCE
    if extension.startswith("."):
        extension = extension[1:]
        
    return [f for f in glob(startpath + f"**/*.{extension}", recursive=True)]

def run_full_filecheck(extension="*", startpath=None,
                        erasebad=True, redownload=False, 
                        nprocess=4, show_progress=True,
                        **kwargs ):
    """ Look for all file with the given extension recursively starting from `startpath` and checks if the file 
    is usable ok not. This returns the bad files.
    
    Parameters
    ----------

    // Data files
    extension: [string] -optional-
        All the 'file.{}'.format(extension) will be looked at.
        (first '.' ignored such that extension='.fits' is equivalent to extension='fits')

    startpath: [None or path] -optional-
        From which directory does this start to look at.
        If None: $ZTFDATA (stored as io.LOCALSOURCE) will be used.

    // Check options

    erasebad: [bool] -optional-
        Do you want to remove from your local directory the corrupted files ?
        
    redownload: [bool] -optional-
        Shall corrupted file be automatically re downloaded ?
        (Only works for IRSA files ('/sci/','/raw/', '/ref/', '/cal/')

    nprocess: [int] -optional-
        Number of paralell processing

    show_progress: [bool] -optional-
        Do you want to show the progress bar ?
        
    Returns
    -------
    list of corrupted/bad files (might already be removed, see erasebad)

    """
    all_ztffiles = get_localfiles(extension=extension, startpath=startpath)
    print(f"{len(all_ztffiles)} files to check")
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        badfiles = test_files(all_ztffiles, erasebad=erasebad, nprocess=nprocess,
                             show_progress=show_progress, 
                             redownload=redownload, **kwargs)
        
    return badfiles
    
def test_files(filename, erasebad=True, nprocess=1, show_progress=True,
                   redownload=False, verbose=False, **kwargs ):
    """ 
    
    Parameters
    ----------
    filename: [fiulepath or list of]
        File(s) to be checked. 

    erasebad: [bool] -optional-
        Do you want to remove from your local directory the corrupted files ?
        
    redownload: [bool] -optional-
        Shall corrupted file be automatically re downloaded ?
        (Only works for IRSA files ('/sci/','/raw/', '/ref/', '/cal/')

    nprocess: [int] -optional-
        Number of paralell processing

    show_progress: [bool] -optional-
        Do you want to show the progress bar ?
               
    Returns
    -------
    list of corrupted/bad files (might already be removed, see erasebad)

    """
    if nprocess is None:
        nprocess = 1
    elif nprocess<1:
        raise ValueError("nprocess must 1 or higher (None means 1)")

    filename = np.atleast_1d(filename)
    
    if nprocess == 1:
        fileissue = [f for f in filename if not _test_file_(f, erasebad=erasebad,
                                                            redownload=redownload,
                                                            **kwargs)]
    else:
        import multiprocessing
        if show_progress:
            from astropy.utils.console import ProgressBar
            bar = ProgressBar( len(filename), ipython_widget=is_running_from_notebook())
        else:
            bar = None

        erasebad_   = [erasebad]*len(filename)
        fileissue = []
        
        with multiprocessing.Pool(nprocess) as p:
            # Da Loop
            for j, isgood in enumerate( p.imap(_test_file_multiprocess_, zip(filename, erasebad_)) ):
                if bar is not None:
                    bar.update(j)
                if not isgood:
                    fileissue.append(filename[j])
                    
            if bar is not None:
                bar.update( len(filename) )
                
    if len(fileissue) >0:
        warnings.warn(f"{len(fileissue)} file failed")
        if redownload:
            from .buildurl import _localsource_to_source_
            to_download_urls, locations = np.asarray([_localsource_to_source_(filename)
                                                        for filename in fileissue]).T
            source_to_dl = ["irsa"]
            for source in source_to_dl:
                source_dl = np.in1d(locations, [source])
                print(f"Downloading {len(source_dl[source_dl])} files from {source}")
                download_url(np.asarray(to_download_urls)[source_dl], np.asarray(fileissue)[source_dl],
                                 show_progress=show_progress, verbose=verbose,
                                 overwrite=True, nprocess=nprocess, cookies=get_cookie(*_load_id_(source)),
                         **kwargs)
            for source_ in np.unique(locations):
                if source_ is not None and source_ not in source_to_dl:
                    warnings.warn(f"files from {source_} have not downloaded (not implemented).")
                
        return fileissue
        

def _test_file_multiprocess_(args):
    """ """
    filename, erasebad = args
    return _test_file_(filename, erasebad=erasebad, fromdl=False, redownload=False)

def _are_fitsfiles_bad_(filenames, test_exist=True):
    """ """
    return [_is_fitsfile_bad_(f_, test_exist=test_exist) for f_ in filenames]

def _is_fitsfile_bad_(filename, test_exist=True):
    """ """
    if not os.path.isfile(filename):
        return test_exist
    
    try:
        _ = fits.getdata(filename)
        return False
    except:
        return True

def _is_textfile_bad_(filename):
    """ """
    try:
        _ = open(filename).read().splitlines()
        return False
    except:
        return True
    
def _test_file_(filename, erasebad=True, fromdl=False,
                redownload=False, verbose=True):
    """ """
    propissue = dict(erasebad=erasebad, fromdl=fromdl, redownload=redownload, verbose=verbose)
    # Fits file
    if ".fits" in filename:
        if not hash_for_file_exists(filename):
            try:
                _ = fits.getdata(filename)
                calculate_and_write_hash(filename)
            except FileNotFoundError:
                warnings.warn(f"[Errno 2] No such file or directory: {filename}")
            except:
                _fileissue_(filename, **propissue)
                return False
        
    # txt file        
    elif ".txt" in filename:
        if not hash_for_file_exists(filename):
            try:
                _ = open(filename).read().splitlines()
                calculate_and_write_hash(filename)
            except FileNotFoundError:
                warnings.warn(f"[Errno 2] No such file or directory: {filename}")
            except:
                _fileissue_(filename, **propissue)
                return False
        
    # other extensions        
    else:
        warnings.warn("no file testing made for .%s files"%filename.split(".")[-1])
            
    return True

def _fileissue_(filename, erasebad=True, fromdl=False, redownload=False, verbose=True):
    """ """
    if fromdl:
        warnings.warn(f"Download failed {filename} seems corrupted (cannot open)")
    else:
        warnings.warn(f"cannot open file {filename}")
        
    if erasebad:
        warnings.warn(f"removing {filename}")
        os.remove(filename)
    else:
        warnings.warn(f"{filename} NOT ERASED")

    if redownload:
        from .buildurl import _localsource_to_source_
        url_to_dl, location = _localsource_to_source_(filename)
        if url_to_dl is not None:
            download_single_url(url_to_dl, fileout=filename, overwrite=True, verbose=verbose,
                                cookies = get_cookie(*_load_id_(location))
                                )
        else:
            warnings.warn("No url to donwload, redownload ignored")
    
# ================= #
#   Logging Tools   #
# ================= #
def get_cookie(username, password):
    """Get a cookie from the IPAC login service

    Parameters
    ----------
    username: [str]
        The IPAC account username
    password: [str]
        The IPAC account password
    """
    url = "%s?josso_cmd=login&josso_username=%s&josso_password=%s" % (LOGIN_URL, username, password)
    return requests.get(url).cookies


def _download_(args):
    """ To be used within _ZTFDownloader_.download_data() 
    url, fileout,overwrite,verbose = args
    """
    url, fileout,  overwrite, verbose, wait = args
    download_single_url(url, fileout=fileout, overwrite=overwrite, verbose=verbose, wait=wait)
    
def download_url(to_download_urls, download_location,
                show_progress = True, verbose=True, wait=None,
                overwrite=False, nprocess=None, cookies=None,
                client=None,
                **kwargs):
    """ """
    #
    # - Dask Client
    if client is not None:
        from dask import delayed
        d_download = [delayed(download_single_url)(url, fileout=fileout, show_progress=False,
                                    overwrite=overwrite, verbose=False, wait=wait,
                                    cookies=cookies, **kwargs)
                    for url, fileout in zip(to_download_urls, download_location)]
        return client.compute(d_download)
    
    #
    # - MultiProcessing (or not)
    if nprocess is None:
        nprocess = 1
    elif nprocess<1:
        raise ValueError("nprocess must 1 or higher (None means 1)")

    
    if nprocess == 1:
        # Single processing
        if verbose:
            warnings.warn("No parallel downloading")
        for url, fileout in zip(to_download_urls, download_location):
            download_single_url(url,fileout=fileout, show_progress=show_progress,
                                    overwrite=overwrite, verbose=verbose, cookies=cookies,
                                    wait=wait,
                                    **kwargs)
    else:
        # Multi processing
        import multiprocessing
        if show_progress:
            from astropy.utils.console import ProgressBar
            bar = ProgressBar( len(to_download_urls), ipython_widget=is_running_from_notebook())
        else:
            bar = None
                
        if verbose:
            warnings.warn("parallel downloading ; asking for %d processes"%nprocess)

        # Passing arguments
        overwrite_ = [overwrite]*len(to_download_urls)
        verbose_   = [verbose]*len(to_download_urls)
        wait_ = [wait]*len(to_download_urls)
        with multiprocessing.Pool(nprocess) as p:
            # Da Loop
            for j, result in enumerate( p.imap_unordered(_download_, zip(to_download_urls,
                                                                    download_location,
                                                                 overwrite_, verbose_, wait_))):
                if bar is not None:
                    bar.update(j)
                    
            if bar is not None:
                bar.update( len(to_download_urls) )

            
def download_single_url(url, fileout=None, 
                        overwrite=False, verbose=True, cookies=None,
                        show_progress=True, chunk=1024,
                        wait=None, randomize_wait=True,
                        filecheck=True, erasebad=True,
                        **kwargs):
    """ Download the url target using requests.get.
    the data is returned (if fileout is None) or stored in `fileout`
    """
    if wait is not None:
        waiting = wait if not randomize_wait else np.random.uniform(0,wait)
        time.sleep(waiting)
        
    if fileout is not None and not overwrite and os.path.isfile( fileout ):
        if verbose:
            warnings.warn("%s already exists: skipped"%fileout)
        return
    else:
        if verbose and fileout:
            warnings.warn("downloading %s to %s"%(url,fileout))

    # = Password and Username
    if cookies is None:
        cookies = get_cookie(*_load_id_("irsa"))

    # - requests options 
    download_prop = dict(cookies=cookies, stream=True)
    for k,v in kwargs.items():
        download_prop[k] = v
        
    if cookies in ["no_cookies"]:
        _ = download_prop.pop("cookies")

    request_fnc = "get" if not "data" in download_prop else "post"
    # = Where should the data be saved?
    if fileout is not None:
        directory = os.path.dirname(fileout)
        if not os.path.exists(directory):
            os.makedirs(directory, exist_ok=True)

    else:
        download_prop["stream"] = False
        return getattr(requests,request_fnc)(url, **download_prop)

    # With Progress bar?
    if not show_progress:
        response = getattr(requests,request_fnc)(url, **download_prop)
        if response.status_code == 200:
            with open(fileout, 'wb') as f:
                for data in response.iter_content(chunk):
                    f.write(data)
    
    else:
        from astropy.utils.console import ProgressBar
        response = getattr(requests,request_fnc)(url, **download_prop)
        if response.status_code == 200:
            chunk_barstep = 500
            f = open(fileout, 'wb')
            with ProgressBar(int(response.headers.get('content-length'))/(chunk_barstep*chunk),
                             ipython_widget=is_running_from_notebook()) as bar:
                for i,data in enumerate(response.iter_content(chunk_size=chunk)):
                    if i%chunk_barstep==0:
                        bar.update()
                    f.write(data)
            f.close()
            calculate_and_write_hash(fileout)

    if filecheck:
        _test_file_(fileout, erasebad=erasebad, fromdl=True)


# ============== #
#                #
#  CC-IN2P3      #
#                #
# ============== #
class CCIN2P3( object ):
    """ """
    def __init__(self, auth=None, connect=True):
        """ """
        if self.running_at_cc:
            self._connected = True
        else:
            self.load_ssh(auth=auth)
            self._connected = False
            if connect:
                self.connect(auth=auth)

    def load_ssh(self, auth=None):
        from paramiko import SSHClient
        self._auth = auth
        self._ssh = SSHClient()
        self._ssh.load_system_host_keys()

    def connect(self, auth=None):
        """ """
        if auth is None:
            auth = _load_id_("ccin2p3")

        username, password= auth
        try:
            self._ssh.connect('cca.in2p3.fr', username=username, password=password)
        except:
            raise IOError("Cannot connect to cca.in2p3.fr with given authentification")
        
        self._connected = True

    @classmethod
    def scp(cls, fromfile, tofile, auth=None):
        """ """
        if fromfile.startswith(CCIN2P3_SOURCE):
            method = "scp_get"
        elif tofile.startswith(CCIN2P3_SOURCE):
            method = "scp_put"
        else:
            raise ValueError(f"None of fromfile or tofile stars with {CCIN2P3_SOURCE}. Cannot use scp(), see scp_get or scp_put")

        this = cls(auth=auth, connect=True)
        return getattr(this, method)(fromfile, tofile)
    
    def scp_get(self, remotefile, localfile, auth=None, overwrite=False):
        """ """
        from scp import SCPClient
        if not self._connected:
            self.connect(auth)

        directory = os.path.dirname(localfile)
        if not os.path.exists(directory):
            warnings.warn(f"scp_get(): creating {directory}")
            os.makedirs(directory, exist_ok=True)

        with SCPClient(self.ssh.get_transport()) as scp:
            scp.get(remotefile, localfile)
            
    def scp_put(self, localfile, remotefile, auth=None):
        """ """
        from scp import SCPClient        
        if not self._connected:
            self.connect(auth)

        with SCPClient(self.ssh.get_transport()) as scp:
            scp.put(localfile,remotefile)

    def query_catalog(self, ra, dec, radius, catname="gaia", depth=7, **kwargs):
        """ query catalog ; works only when logged at the CCIN2P3"""
        if not self.running_at_cc:
            raise IOError("Only works if running from the ccin2p3")
        
        import os
        from htmcatalog import htmquery
        LSST_REFCAT_DIR = "/sps/lsst/datasets/refcats/htm/v1"
        KNOW_REFCAT = {"gaiadr1":"gaia_DR1_v1",
                       "gaiadr2":"gaia_dr2_20190808",
                       "ps1dr1":"ps1_pv3_3pi_20170110",
                       "sdssdr9":"sdss-dr9-fink-v5b",
                      }
        KNOW_REFCAT["gaia"] = KNOW_REFCAT["gaiadr2"]
        KNOW_REFCAT["ps1"] = KNOW_REFCAT["ps1dr1"]
        if catname not in KNOW_REFCAT:
            raise ValueError(f"unknown catalog {catname}. Aviability: "+", ".join(list(KNOW_REFCAT.keys())))


        hq = htmquery.HTMQuery(depth, os.path.join(LSST_REFCAT_DIR, KNOW_REFCAT[catname]))
        return hq.fetch_cat(ra, dec, radius, **kwargs)
                                   

    # ============= #
    #  Properties   #
    # ============= #
    @property
    def ssh(self):
        """ """
        return self._ssh
    
    @property
    def running_at_cc(self):
        """ """
        hostname = os.uname()[1]
        return "cca" in hostname or "ccwige" in hostname


# =============== #
#                 #
#  HASH tools     #
#                 #
# =============== #


        
def calculate_hash(fname):
    """ """
    f = open(fname, 'rb')
    hash_md5 = hashlib.md5()
    for chunk in iter(lambda: f.read(4096), b""):
        hash_md5.update(chunk)
    hexdigest = hash_md5.hexdigest()
    f.close()
    return hexdigest

def calculate_and_write_hash(fname):
    """ """
    f = open(fname, 'rb')
    hash_md5 = hashlib.md5()
    for chunk in iter(lambda: f.read(4096), b""):
        hash_md5.update(chunk)
    hexdigest = hash_md5.hexdigest()
    f.close()
    f_hash = open(f"{fname}.md5", 'w')
    f_hash.write(hexdigest)
    f_hash.close()

def read_hash(fname):
    """ """
    f_hash = open(f"{fname}.md5", 'r')
    hash_md5 = f_hash.read()
    return hash_md5

def compare_hash(fname):
    """ """
    f_hash = open(hash_fname, 'r')
    hash_md5_read = f_hash.read()
    hash_md5_calculated = calculate_hash(fname)
    if hash_md5_read == hash_md5_calculated:
        return True
    else:
        return False

def hash_for_file_exists(fname):
    """ """
    return os.path.exists(f"{fname}.md5")



#! /usr/bin/env python
#

""" Combine MetaSearch and MetaURL to get data from IRSA """

import os
import numpy as np
from .metasearch import download_metadata, _test_kind_
from . import buildurl, ztftable
import warnings


# This enables multiprocess downloading
from . import io



# Combining metadata with buildurl
def metatable_to_url(metatable, datakind='sci', suffix=None, source=None, **kwargs):
    """generic method to build the url/fullpath or the requested data.
        This method is based on the `builurl.py` module of ztfquery.
        
        Parameters
        ----------
        suffix: [string] -optional-
            What kind of data do you want? 
            Here is the list of available options depending the image kind:
        
            # Science image (kind="sci"):
            - sciimg.fits (primary science image) # (default)
            - mskimg.fits (bit-mask image)
            - psfcat.fits (PSF-fit photometry catalog)
            - sexcat.fits (nested-aperture photometry catalog)
            - sciimgdao.psf (spatially varying PSF estimate in DAOPhot's lookup table format)
            - sciimgdaopsfcent.fits (PSF estimate at science image center as a FITS image)
            - sciimlog.txt (log output from instrumental calibration pipeline)
            - scimrefdiffimg.fits.fz (difference image: science minus reference; fpack-compressed)
            - diffimgpsf.fits (PSF estimate for difference image as a FITS image)
            - diffimlog.txt (log output from image subtraction and extraction pipeline)
            - log.txt (overall system summary log from realtime pipeline)
            
            # Reference image (kind="ref"):
            -log.txt
            -refcov.fits
            -refimg.fits # (default)
            -refimlog.txt
            -refpsfcat.fits
            -refsexcat.fits
            -refunc.fits

            # Raw images (kind="raw")
            No Choice so suffix is ignored for raw data
            
            # Calibration (kind="cal")
            - None (#default) returns `caltype`.fits
            - log:            returns `caltype`log.txt
            - unc:            returns `caltype`unc.fits

        // if queried metadata is for kind calibration

        """
    if datakind in ['sci', "raw"]:    
        filtercode,imgtypecode  = np.asarray(metatable[["filtercode","imgtypecode"]
                                                                    ].values.T, dtype="str")
        paddedfield      = np.asarray(["%06d"%f for f in metatable["field"].values],
                                          dtype="str")
        paddedccdid      = np.asarray(["%02d"%f for f in metatable["ccdid"].values],
                                              dtype="str")
        year, month, day, fracday = np.asarray([[l[:4],l[4:6],l[6:8],l[8:]]
                                for l in np.asarray(metatable["filefracday"].values,
                                              dtype="str") ]).T
  
        if datakind in ['sci']:
            qid  = np.asarray(metatable["qid"], dtype="str")
                # LIST of URL to download [SCIENCE]
            return  [buildurl.science_path(year_, month_, day_, fracday_, paddedfield_,
                                filtercode_, paddedccdid_, qid_,
                                imgtypecode=imgtypecode_, suffix=suffix, source=source, **kwargs)
                        for year_, month_, day_, fracday_, paddedfield_, filtercode_,
                        paddedccdid_, qid_, imgtypecode_
                        in zip(year, month, day, fracday, paddedfield, filtercode,
                               paddedccdid, qid, imgtypecode)]
        else:
            # LIST of URL to download [RAW]
            return  [buildurl.raw_path(year_, month_, day_, fracday_,
                                           paddedfield_ if imgtypecode_ != "f" else '000000',  # because sometime they do have a field, why is that ?,
                              filtercode_, paddedccdid_, 
                              imgtypecode=imgtypecode_, source=source, **kwargs)
                        for year_, month_, day_, fracday_, paddedfield_, filtercode_,
                        paddedccdid_,  imgtypecode_
                        in zip(year, month, day, fracday, paddedfield, filtercode,
                                   paddedccdid, imgtypecode)]
    # CALIBRATION
    elif datakind in ['cal']:
        year, month, day = np.asarray([[l[:4],l[4:6],l[6:]]
                                for l in np.asarray(metatable["nightdate"].values,
                                                        dtype="str") ]).T
        paddedccdid      = np.asarray(["%02d"%f for f in metatable["ccdid"].values],
                                              dtype="str")
        filtercode, qid,caltype  = np.asarray(metatable[["filtercode", "qid","caltype"]].values.T,
                                                      dtype="str")
        filtercode[filtercode=="0"] = "00" # such that it is what IRSA expects.
            # list of url to download [CAL]
        return  [buildurl.calibration_path(caltype_,
                                            year_, month_, day_,
                                            filtercode_, paddedccdid_, qid_,
                                            suffix=suffix, source=source, **kwargs)
                            for caltype_, year_, month_, day_, filtercode_, paddedccdid_, qid_
                     in zip(caltype,year, month, day,filtercode, paddedccdid, qid) ]
    # PIXELS
    elif datakind in ['ref']:
        paddedfield      = np.asarray(["%06d"%f for f in metatable["field"].values], dtype="str")
        paddedccdid      = np.asarray(["%02d"%f for f in metatable["ccdid"].values], dtype="str")
        filtercode, qid  = np.asarray(metatable[["filtercode", "qid"]].values.T, dtype="str")
 
        return [buildurl.reference_path( paddedfield_,
                                filtercode_, paddedccdid_, qid_,
                                suffix=suffix,
                                fieldprefix=paddedfield_[:3], # This is how it is defined in IRSA
                                source=source, **kwargs)
                        for  paddedfield_, filtercode_, paddedccdid_, qid_
                    in zip(paddedfield, filtercode, paddedccdid, qid)]
        
def guess_kind_from_metatable(df):
    """ Given what is inside the dataframe, this guesses the data kind"""
    if "caltype" in df.columns:
        return "cal"
    if "startobsdate" in df.columns:
        return "ref"
    if "rcid" in df.columns:
        return "sci"
    if "expid" in df.columns:
        return "raw"
    raise TypeError("The given dataframe is not a ZTF IRSA metatable")

#############################
#                           #
#   Main Query Tools        #
#                           #
#############################
    
class _ZTFDownloader_( object ):
    """ Virtual class that enable to download consistently ZTF data. 
    To use it, you need to inherite this and implement get_data_path()
    such that this method returns fullpath to the data given 
    `suffix` and `source` arguments.
    """
    def get_data_path(self, suffix=None, source=""):
        """ generic method to build the url/fullpath or the requested data.
        This method is based on the `builurl.py` module of ztfquery.

        **This is a virtual empty function ; inheriting class must implemented This**
        """
        raise NotImplementedError("the get_data_path() method must be implemented. ")

    
    # Generic that should automatically work as long as get_data_path is defined.
    def download_data(self, suffix=None, source="IRSA", indexes=None,
                     download_dir=None, show_progress = True, verbose=True,
                     nodl = False, overwrite=False, nprocess=None, auth=None,
                     filecheck=True, erasebad=True, redownload=True,
                     ignore_warnings=False, client=None, 
                     **kwargs):
        """ 
        Parameters
        ----------
        suffix: [string] -optional-
            What kind of data do you want? 
            Here is the list of available options depending on you image kind:
        
            # Science image (kind="sci"):
            - sciimg.fits (primary science image) # (default)
            - mskimg.fits (bit-mask image)
            - psfcat.fits (PSF-fit photometry catalog)
            - sexcat.fits (nested-aperture photometry catalog)
            - sciimgdao.psf (spatially varying PSF estimate in DAOPhot's lookup table format)
            - sciimgdaopsfcent.fits (PSF estimate at science image center as a FITS image)
            - sciimlog.txt (log output from instrumental calibration pipeline)
            - scimrefdiffimg.fits.fz (difference image: science minus reference; fpack-compressed)
            - diffimgpsf.fits (PSF estimate for difference image as a FITS image)
            - diffimlog.txt (log output from image subtraction and extraction pipeline)
            - log.txt (overall system summary log from realtime pipeline)
            
            # Reference image (kind="ref"):
            -log.txt
            -refcov.fits
            -refimg.fits # (default)
            -refimlog.txt
            -refpsfcat.fits
            -refsexcat.fits
            -refunc.fits

            # Raw images (kind="raw")
            No Choice. Suffix is ignored for raw data
            
            # Calibration (kind="cal")
            - None (#default) returns `caltype`.fits
            - log:            returns `caltype`log.txt
            - unc:            returns `caltype`unc.fits

        download_dir: [string] -optional-
            Directory where the file should be downloaded.
            If th
            
        overwrite: [bool] -optional-
            Check if the requested data already exist in the target download directory. 
            If so, this will skip the download except if overwrite is set to True.

        nprocess: [None/int] -optional-
            Number of parallel downloading you want to do. 
            If None, it will be set to 1 and will not use multiprocess

        auth: [str, str] -optional-
            [username, password] of you IRSA account.
            If used, information stored in ~/.ztfquery will be ignored.
        
        //  File check

        filecheck: [bool] -optional-
            Shall this check if the downloaded data actually works ?

        erasebad: [bool] -optional-
            Do you want to remove from your local directory the corrupted files ?

        redownload: [bool] -optional-
            Shall corrupted file be automatically re downloaded ?
            (Only works for IRSA files ('/sci/','/raw/', '/ref/', '/cal/')

        """
        # login
        if auth is not None:
            from .io import get_cookie
            cookie = get_cookie(*auth)
        else:
            cookie = None
            
        # Data Structure
        self._relative_data_path = self.get_data_path(suffix=suffix, source="None",
                                                      indexes=indexes, **kwargs)
        
        # The IRSA location
        self.to_download_urls    = [os.path.join(buildurl._source_to_location_(source),d_)
                                     for d_ in self._relative_data_path]
        # Where do you want them?
        if download_dir is None: # Local IRSA structure
            self.download_location   = [os.path.join(buildurl._source_to_location_("local"), d_)
                                        for d_ in self._relative_data_path]
        else:
            self.download_location   = [os.path.join(download_dir,"%s"%(d_.split("/")[-1]))
                                        for d_ in self._relative_data_path]
            
        if nodl:
            return self.to_download_urls, self.download_location
        
        # Actual Download
        with warnings.catch_warnings():
            if ignore_warnings:
                warnings.simplefilter("ignore")
            delayed_ = io.download_url(self.to_download_urls, self.download_location,
                        show_progress = show_progress, verbose=verbose,
                        overwrite=overwrite, nprocess=nprocess, cookies=cookie,
                        filecheck=False, client=client)
            print("delayed")
            if client is not None:
                return delayed_
            
        if filecheck:
            with warnings.catch_warnings():
                if ignore_warnings:
                    warnings.simplefilter("ignore")
                fileissue = io.test_files(self.download_location, erasebad=erasebad,
                                              redownload=True, nprocess=nprocess)
                
            if fileissue is not None and len(fileissue) > 0:
                warnings.warn("%d file failed (returned)"%len(fileissue))
                
                
    # --------- #
    #  GETTER   #
    # --------- #
    def get_data(self, suffix=None, indexes=None, downloadit=True, check_suffix=True,
                     dlfrom="irsa", overwrite=False, maxnprocess=4, exist=True,
                     show_progress=True, **kwargs):
        """ high-level method to get the data you want.
        It will look for the data locally and download the one that are missing.
        Set exist = False to have the expected list of data.

        suffix could be a list of suffix, e.g. ["sciimg.fits", "mskimg.fits"]
        
        Parameters
        ----------

        Returns
        -------
        list of path.
        """
        return io.get_file(self.get_data_path(indexes=indexes), suffix=suffix, exist=exist,
                               downloadit=downloadit, show_progress=show_progress, dlfrom=dlfrom,
                               maxnprocess=maxnprocess, check_suffix=check_suffix,
                               overwrite=overwrite, **kwargs)
    
    
    def get_local_data(self, suffix=None, exists=True, filecheck=True, indexes=None,
                           badfiles=False, source="local",
                           ignore_warnings=False, fileout=None, check_suffix=True, **kwargs):
        """ the lists of files stored in your local copy of the ztf database.
        [This methods uses the get_data_path() method assuming source='local']

        Parameters
        ----------
        suffix: [string] -optional-
            What kind of data do you want? 
            Here is the list of available options depending on you image kind:
        
            # Science image (kind="sci"):
            - sciimg.fits (primary science image) # (default)
            - mskimg.fits (bit-mask image)
            - psfcat.fits (PSF-fit photometry catalog)
            - sexcat.fits (nested-aperture photometry catalog)
            - sciimgdao.psf (spatially varying PSF estimate in DAOPhot's lookup table format)
            - sciimgdaopsfcent.fits (PSF estimate at science image center as a FITS image)
            - sciimlog.txt (log output from instrumental calibration pipeline)
            - scimrefdiffimg.fits.fz (difference image: science minus reference; fpack-compressed)
            - diffimgpsf.fits (PSF estimate for difference image as a FITS image)
            - diffimlog.txt (log output from image subtraction and extraction pipeline)
            - log.txt (overall system summary log from realtime pipeline)
            
            # Reference image (kind="ref"):
            -log.txt
            -refcov.fits
            -refimg.fits # (default)
            -refimlog.txt
            -refpsfcat.fits
            -refsexcat.fits
            -refunc.fits

            # Raw images (kind="raw")
            No Choice so suffix is ignored for raw data
            
            # Calibration (kind="cal")
            - None (#default) returns `caltype`.fits
            - log:            returns `caltype`log.txt
            - unc:            returns `caltype`unc.fits

        exists: [bool] -optional-
            returns only the file that exists in your computer. 
            If false, this will return the expected path of the requested data, 
            even though they might not exist.


        source: [str] -optional-
            Where are the data stored ? 
            default is 'local' this will use you $ZTFDATA path. 
            You can also directly provide here where the IRSA data structure starts.
        Returns
        -------
        list
        """
        files = self.get_data_path(suffix=suffix, source=source, indexes=indexes, check_suffix=check_suffix)
        if not exists:
            return files

        localfile = [f for f in files if os.path.isfile( f )]
        
        if filecheck or badfiles or kwargs.get("redownload",False):
            with warnings.catch_warnings():
                if ignore_warnings:
                    warnings.simplefilter("ignore")
                badfiles_ = io.test_files(localfile, erasebad=False, **kwargs)
                
            if badfiles:
                return badfiles_
            elif badfiles_ is not None and len(badfiles_)>1:
                return [f for f in localfile if f not in badfiles_]

        if fileout is not None:
            with open(fileout, "w") as f_:
                [f_.write(l+"\n") for l in localfile]
                
        return localfile

    def get_missing_data_index(self, suffix=None, which="any"):
        """ 
        Parameters
        ----------
        suffix: [string] -optional-
            What kind of data do you want? 
            Here is the list of available options depending on you image kind:
        
            # Science image (kind="sci"):
            - sciimg.fits (primary science image) # (default)
            - mskimg.fits (bit-mask image)
            - psfcat.fits (PSF-fit photometry catalog)
            - sexcat.fits (nested-aperture photometry catalog)
            - sciimgdao.psf (spatially varying PSF estimate in DAOPhot's lookup table format)
            - sciimgdaopsfcent.fits (PSF estimate at science image center as a FITS image)
            - sciimlog.txt (log output from instrumental calibration pipeline)
            - scimrefdiffimg.fits.fz (difference image: science minus reference; fpack-compressed)
            - diffimgpsf.fits (PSF estimate for difference image as a FITS image)
            - diffimlog.txt (log output from image subtraction and extraction pipeline)
            - log.txt (overall system summary log from realtime pipeline)
            
            # Reference image (kind="ref"):
            -log.txt
            -refcov.fits
            -refimg.fits # (default)
            -refimlog.txt
            -refpsfcat.fits
            -refsexcat.fits
            -refunc.fits

            # Raw images (kind="raw")
            No Choice so suffix is ignored for raw data
            
            # Calibration (kind="cal")
            - None (#default) returns `caltype`.fits
            - log:            returns `caltype`log.txt
            - unc:            returns `caltype`unc.fits


        which: [string] -optional-
            Which missing data are you looking for:
            - any/all: missing because bad local files or because not downloaded yet.
            - bad/corrupted: missing because the local files are corrupted
            - notdl/dl/nodl: missing because not downloaded. 

        Returns
        -------
        list of self.metadata indexes
        """
        return self.get_local_metatable(suffix=suffix, which=which, invert=True).index
        
    def purge_corrupted_local_data(self, suffix=None, erasebad=False, redownload=False,
                                       indexes=None, verbose=True,
                                       nprocess=4, **kwargs):
        """ 
        Parameters
        ----------
        suffix: [string] -optional-
            What kind of data do you want? 
            Here is the list of available options depending on you image kind:
        
            # Science image (kind="sci"):
            - sciimg.fits (primary science image) # (default)
            - mskimg.fits (bit-mask image)
            - psfcat.fits (PSF-fit photometry catalog)
            - sexcat.fits (nested-aperture photometry catalog)
            - sciimgdao.psf (spatially varying PSF estimate in DAOPhot's lookup table format)
            - sciimgdaopsfcent.fits (PSF estimate at science image center as a FITS image)
            - sciimlog.txt (log output from instrumental calibration pipeline)
            - scimrefdiffimg.fits.fz (difference image: science minus reference; fpack-compressed)
            - diffimgpsf.fits (PSF estimate for difference image as a FITS image)
            - diffimlog.txt (log output from image subtraction and extraction pipeline)
            - log.txt (overall system summary log from realtime pipeline)
            
            # Reference image (kind="ref"):
            -log.txt
            -refcov.fits
            -refimg.fits # (default)
            -refimlog.txt
            -refpsfcat.fits
            -refsexcat.fits
            -refunc.fits

            # Raw images (kind="raw")
            No Choice so suffix is ignored for raw data
            
            # Calibration (kind="cal")
            - None (#default) returns `caltype`.fits
            - log:            returns `caltype`log.txt
            - unc:            returns `caltype`unc.fits


        erasebad: [bool] -optional-
            Do you want to remove from your local directory the corrupted files ?
        
        redownload: [bool] -optional-
            Shall corrupted file be automatically re downloaded ?
            (Only works for IRSA files ('/sci/','/raw/', '/ref/', '/cal/')

        Returns
        -------
        list of files
        """
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            badfiles = self.get_local_data(suffix=suffix, exists=True, filecheck=True,
                                            indexes=indexes, badfiles=True,
                                            redownload=redownload, nprocess=nprocess, **kwargs)

        if badfiles is None or len(badfiles)==0:
            return None
        
        if erasebad and not redownload:
            if verbose:
                print("Removing %d files"%len(badfiles))
            for file_ in badfiles:
                os.remove(file_)
        elif redownload:
            if verbose:
                print("%d files have been redownloaded"%len(badfiles))
        else:
            if verbose:
                print("%d files to be removed (run this with erasebad=True)"%len(badfiles))
                
        return badfiles

    
class ZTFQuery( ztftable._ZTFTable_, _ZTFDownloader_ ):
    """ """
    def __init__(self, metatable=None, kind=None):
        """ """
        if metatable is not None:
            self.set_metatable(metatable, kind)

    @classmethod
    def from_metafile(cls, metafile, **kwargs):
        """ Loads a ZTFQuery instance from a metatable file. """
        import pandas
        metatable = pandas.read_csv(metafile, **{**{"index_col":0},**kwargs})
        kind = guess_kind_from_metatable(metatable)
        return cls(metatable, kind)

    @classmethod
    def from_metaquery(cls, kind="sci", radec=None, size=None, caltype=None,
                           sql_query=None, auth=None, **kwargs):
        """ Loads a ZTFQuery instance and runs load_metadata with the given input """
        this = cls()
        this.load_metadata(kind=kind, radec=radec, size=size, caltype=caltype,
                               sql_query=sql_query, auth=auth, **kwargs)
        return this
    
    def set_metatable(self, metatable, kind):
        """ directly provide the metatable of interest"""
        self._metatable = metatable
        if kind not in ["sci","raw","cal","ref"]:
            raise ValueError(f"The given metatable should be of kind sci/raw/cal/ref, {kind} given")
        
        self._datakind = kind
        
    # ------------ #
    #  DOWNLOADER  #
    # ------------ #
    def load_metadata(self, kind="sci",
                        radec=None, size=None, mcen=None,
                        caltype=None,
                        sql_query=None, auth=None, **kwargs):
        """ Querying for the metadata information that enables to reconstruct 
        the URL to access the data.
        
        [This methods uses the .metasearch library, which is python wrapper of the the IRSA web API
        see https://irsa.ipac.caltech.edu/docs/program_interface/ztf_api.html]
        
        Parameters
        ----------
        kind: [str] -optional-
            What kind of data are you looking for:
            - sci : Science Exposures
            - raw : Raw Data
            - ref : Reference Images
            - cal : Bias or High Frequency Flat 
            any other entry will raise a ValueError

        // Generic Query

        sql_query: [None or string] -optional - 
            The where parameter can be set to a 'SQL WHERE' clause, with some restrictions. 
            [https://en.wikipedia.org/wiki/Where_(SQL)]
            Notably, function calls and sub-queries are not supported. 
            You can use AND, OR, NOT, IN, BETWEEN, LIKE, IS, 
            the usual arithmetic and comparison operators, and literal values.
            Note that the where parameter is required in the absence of POS (a spatial constraint).
            WHERE clauses should be URL encoded 
            [https://en.wikipedia.org/wiki/Query_string#URL_encoding].
            for instance  SPACE is encoded as '+' or "%20".
            If entry must be equal to a string, use `entry='str'` (with the quotes)
            Examples:
                get all the science field 600
                ```field=600```
                get all the science field 600 and having an airmass greater than 2
                ```field=600+AND+airmass>2```
                get all the science field 600 and having an airmass greater than 2 with a quadran ID been 1 or 3
                ```field=600+AND+airmass>2+AND+qid+IN+(1,3)```
                get observation taken since the 1st of Feb 2018 (julian date 2458150.5) with an airmass > 3
                ```airmass>3+AND+obsjd>2458150.5```
    
        // If not Calibration //

        ra,dec: [float/str]
            ICRS right ascension and declination in decimal degrees.
            It identifies the point which returned images must contain, or the center of the search region.

        size: [float/str/None] -optional-
            It consists of one or two (comma separated) values in decimal degrees. 
            (With POS=ra,dec)
            The first value is taken to be the full-width of the search region along the east axis at POS, 
            and the second is taken to be the full-height along the north axis. 
            Taken together, POS and SIZE define a convex spherical polygon on the sky with great circle edges - the search region. 
            During a query, this region is compared against the convex spherical polygons formed by connecting 
            the 4 corners of each image in a data-set to determine which images should be returned.
            
            If only one size value is specified, it is used as both the full-width and full-height.
            Negative sizes are illegal, and a width and height of zero indicate that the search region is a point.

        mcen: [bool] -optional-
            [If the size parameter is specified and non-zero, the mcen parameter is ignored] 

            The mcen parameter indicates that only the most centered image/image set 
            (with respect to POS) should be returned, rather than all images/image sets containing POS. 

        // If Raw //
        INFO: 
            sql_query has a entry to select the kind of data you are looking for:
            - 'imgtypecode'- 
            Currently, the possible values for "imgtypecode" pertaining to raw image data files are:
                o = science (on-sky) 
                b = bias calibration image
                d = dark calibration image
                f = dome/screen flatfield calibration image

        // If Calibration //
        caltype: [strin]
            which calibration type? 'bias' or 'hifreqflat'
            This classification will be added to the sql_query (caltype=`caltype`) 
            except if the sql_query already contains it. 
            If None, this will be ignored 
    
        """
        _test_kind_(kind)
        if auth is not None:
            from .io import get_cookie
            kwargs["cookies"] = get_cookie(*auth)
        
        if kind not in ['cal']:
            # python3 -> self.metaquery = download_metadata(**{**locals(),**kwargs})
            self.metaquery = download_metadata(kind=kind, radec=radec, size=size,
                                                   mcen=mcen, sql_query=sql_query, **kwargs)
        else:
            for k in ["radec", "size", "mcen"]:
                if locals()[k] is not None: warnings.warn("You are querying 'calibration data', the following entry is ignored: %s"%k)
            if "caltype" not in sql_query and caltype is not None and caltype:
                sql_query = "caltype=%s"%caltype if sql_query is None else sql_query+"+AND+caltype='%s'"%caltype
                
            self.metaquery = download_metadata(kind=kind, sql_query=sql_query, **kwargs)
            
    
    def get_data_path(self, suffix=None, source=None, indexes=None, **kwargs):
        """ generic method to build the url/fullpath or the requested data.
        This method is based on the `builurl.py` module of ztfquery.
        
        Parameters
        ----------
        suffix: [string] -optional-
            What kind of data do you want? 
            Here is the list of available options depending on you image kind:
        
            # Science image (kind="sci"):
            - sciimg.fits (primary science image) # (default)
            - mskimg.fits (bit-mask image)
            - psfcat.fits (PSF-fit photometry catalog)
            - sexcat.fits (nested-aperture photometry catalog)
            - sciimgdao.psf (spatially varying PSF estimate in DAOPhot's lookup table format)
            - sciimgdaopsfcent.fits (PSF estimate at science image center as a FITS image)
            - sciimlog.txt (log output from instrumental calibration pipeline)
            - scimrefdiffimg.fits.fz (difference image: science minus reference; fpack-compressed)
            - diffimgpsf.fits (PSF estimate for difference image as a FITS image)
            - diffimlog.txt (log output from image subtraction and extraction pipeline)
            - log.txt (overall system summary log from realtime pipeline)
            
            # Reference image (kind="ref"):
            -log.txt
            -refcov.fits
            -refimg.fits # (default)
            -refimlog.txt
            -refpsfcat.fits
            -refsexcat.fits
            -refunc.fits

            # Raw images (kind="raw")
            No Choice so suffix is ignored for raw data
            
            # Calibration (kind="cal")
            - None (#default) returns `caltype`.fits
            - log:            returns `caltype`log.txt
            - unc:            returns `caltype`unc.fits
        // if queried metadata is for kind calibration
            
        """
        if len(self.metatable) == 0:
            warnings.warn("No entry associated to the query you made: metatable is empty")
            return []
        if "check_suffix" in kwargs and self.datakind not in ["cal","sci"]:
            _ = kwargs.pop("check_suffix")
            
        return metatable_to_url(self.metatable if indexes is None else self.metatable.loc[np.atleast_1d(indexes)],
                                    datakind=self.datakind, suffix=suffix, source=source, **kwargs)

    def get_local_metatable(self, suffix=None, which="any", invert=False):
        """ 
        Parameters
        ----------
        suffix: [string] -optional-
            What kind of data do you want? 
            Here is the list of available options depending on you image kind:
        
            # Science image (kind="sci"):
            - sciimg.fits (primary science image) # (default)
            - mskimg.fits (bit-mask image)
            - psfcat.fits (PSF-fit photometry catalog)
            - sexcat.fits (nested-aperture photometry catalog)
            - sciimgdao.psf (spatially varying PSF estimate in DAOPhot's lookup table format)
            - sciimgdaopsfcent.fits (PSF estimate at science image center as a FITS image)
            - sciimlog.txt (log output from instrumental calibration pipeline)
            - scimrefdiffimg.fits.fz (difference image: science minus reference; fpack-compressed)
            - diffimgpsf.fits (PSF estimate for difference image as a FITS image)
            - diffimlog.txt (log output from image subtraction and extraction pipeline)
            - log.txt (overall system summary log from realtime pipeline)
            
            # Reference image (kind="ref"):
            -log.txt
            -refcov.fits
            -refimg.fits # (default)
            -refimlog.txt
            -refpsfcat.fits
            -refsexcat.fits
            -refunc.fits

            # Raw images (kind="raw")
            No Choice so suffix is ignored for raw data
            
            # Calibration (kind="cal")
            - None (#default) returns `caltype`.fits
            - log:            returns `caltype`log.txt
            - unc:            returns `caltype`unc.fits


        which: [string] -optional-
            Which missing data are you looking for:
            - any/all: missing because bad local files or because not downloaded yet.
            - bad/corrupted: missing because the local files are corrupted
            - notdl/dl/nodl: missing because not downloaded. 

        invert: [bool] -optional-
            Set True to invert the method, i.e. get the metable of the files *you do not have*

        Returns
        -------
        self.metadata (filtered to match your local data)
        """
        
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            if which in ["all", "any"]:
                all_local = self.get_data_path(suffix ,source="local")
                actual_local = self.get_local_data(suffix, exists=True, filecheck=True)            
            elif which in ["bad", "corrupted"]:
                all_local = self.get_local_data(suffix, exists=True, filecheck=False)
                actual_local = self.get_local_data(suffix, exists=True, filecheck=True)
            elif which in ["dl"]:
                all_local = self.get_data_path(suffix ,source="local")
                actual_local = self.get_local_data(suffix, exists=True, filecheck=False)
            else:
                raise ValueError("cannot parse which %s: any, bad, notdl available"%which)

        flagin = np.in1d(all_local, actual_local)
        return self.metatable[flagin if not invert else ~flagin]
    # =============== #
    #  Properties     #
    # =============== #
    @property
    def datakind(self):
        """ """
        if not hasattr(self,"_datakind"):
            if not hasattr(self, "metaquery"):
                raise AttributeError("metaquery has not been loaded. Run load_metadata(). ")
            return self.metaquery.datakind
        
        return self._datakind
    
    @property
    def metatable(self):
        """ """
        if not hasattr(self,"_metatable"):
            if not hasattr(self, "metaquery"):
                raise AttributeError("No metatable has not been loaded. Run load_metadata(). ")
        
            return self.metaquery.metatable
        
        return self._metatable

    @property
    def data(self):
        """ short cut towards metatable """
        # enables to _ZTFTable_ tricks
        return self.metatable
    
#############################
#                           #
#  Addition Queries         #
#                           #
#############################
_NIGHT_SUMMARY_URL = "http://www.astro.caltech.edu/~tb/ztfops/sky/"
def download_night_summary(night, ztfops_auth = None):
    """ 
    Parameters
    ----------
    night: [string]
        Format: YYYYMMDD like for instance 20180429

    ztfops_auth: [string, string] -optional-
        Provide directly the [username, password] of the ztfops page.
    """
    print("NIGHT SUMMARY IS NOW DEPRECATED, USE skyvision.CompletedLog.from_date('YYY-MM-DD') instead")
    return

def download_allnight_summary(ztfops_auth = None):
    """ 
    Parameters
    ----------
    ztfops_auth: [string, string] -optional-
        Provide directly the [username, password] of the ztfops page.
    """
    print("NIGHT SUMMARY IS NOW DEPRECATED, USE skyvision.CompletedLog.from_date('YYY-MM-DD') instead")
    return




#! /usr/bin/env python
#

"""  
https://irsa.ipac.caltech.edu/docs/program_interface/ztf_api.html
"""

from io import StringIO
    
import requests
import warnings
import numpy as np
from pandas import read_csv


__all__ = ["download_metadata"]

IRSA_DOC = "https://irsa.ipac.caltech.edu/docs/program_interface/ztf_metadata.html"

SEARCH_BASEURL = "https://irsa.ipac.caltech.edu/ibe/search/ztf/products/"

# This might not be up to date, use `get_queriable_field()` to get the latest version
QUERIABLE_FIELD ={"sci":['ra', 'dec', 'infobits',
                         'field', 'ccdid', 'qid', 'rcid', 'fid',
                         'filtercode', 'pid', 'nid', 'expid', 'itid', 'imgtype',
                         'imgtypecode', 'obsdate', 'obsjd', 'exptime', 'filefracday',
                         'seeing', 'airmass', 'moonillf', 'moonesb', 'maglimit', 'crpix1',
                         'crpix2', 'crval1', 'crval2', 'cd11', 'cd12', 'cd21', 'cd22', 'ra1',
                         'dec1', 'ra2', 'dec2', 'ra3', 'dec3', 'ra4', 'dec4', 'ipac_pub_date', 'ipac_gid',
                         'meta_id', 'poly'],
                    "raw":['infobits', 'field', 'ccdid', 'fid', 'filtercode',
                           'nid', 'expid', 'itid', 'imgtype', 'imgtypecode', 'filefracday',
                           'obsdate', 'obsjd', 'exptime', 'ipac_gid', 'seeing', 'airmass',
                           'moonillf', 'moonesb', 'telra', 'teldec', 'ccdrot', 'ipac_pub_date',
                           'ipac_gid', 'meta_id'],
                    "ref":['infobits', 'field', 'ccdid', 'qid', 'rcid', 'fid', 'filtercode',
                            'rfid', 'nframes', 'maglimit', 'startobsdate', 'endobsdate', 'crpix1',
                            'crpix2', 'crval1', 'crval2', 'cd11', 'cd12', 'cd21', 'cd22', 'ra', 'dec',
                            'ra1', 'dec1', 'ra2', 'dec2', 'ra3', 'dec3', 'ra4', 'dec4',
                            'ipac_pub_date', 'ipac_gid', 'meta_id', 'poly'],
                    "cal":['column_name', 'caltype', 'ccdid', 'qid', 'rcid', 'fid', 'filtercode',
                            'cid', 'nightdate', 'startobsdate', 'endobsdate', 'nid', 'nframes',
                            'ipac_pub_date', 'meta_id', 'obsdate_range']
                    }

# INFORMATION
SCIENCE_FIELDS = """
ra	Right ascension of image center	deg	pos.eq.ra		double	0	0
dec	Declination of image center	deg	pos.eq.dec		double	0	0
infobits	Info bit flags				int	0	0
field	ZTF field number				int	0	1
ccdid	CCD number (1..16)				int	0	0
qid	Quadrant ID (1..4)				int	0	0
rcid	Readout-channel ID (0..63)				int	0	0
fid	Filter ID				int	0	0
filtercode	Filter code (abbreviated name)				char	0	0
pid	Science product ID				long	0	1
nid	Day/night ID				int	0	1
expid	Exposure ID				int	0	1
itid	Image type ID				int	0	0
imgtype	Image type name				char	0	0
imgtypecode	Single letter image type code				char	0	0
obsdate	Observation UT date/time				char	0	1
obsjd	Observation Julian day	d			double	0	1
exptime	Exposure time	s			double	0	0
filefracday	Integer (YYYYMMDDdddddd) used in product file names				long	0	0
seeing	Seeing FWHM	arcsec			double	0	0
airmass	Telescope airmass				double	0	0
moonillf	Moon illuminated fraction				double	0	0
moonesb	Moon excess in sky brightness				double	0	0
maglimit	Magnitude limit	mag			double	0	0
crpix1	Reference pixel value for axis 1	pix			double	0	0
crpix2	Reference pixel value for axis 2	pix			double	0	0
crval1	Reference position right ascension at crpix1,crpix2	deg			double	0	0
crval2	Reference position declination at crpix1,crpix2	deg			double	0	0
cd11	CD matrix element 1,1	deg/pix			double	0	0
cd12	CD matrix element 1,2	deg/pix			double	0	0
cd21	CD matrix element 2,1	deg/pix			double	0	0
cd22	CD matrix element 2,2	deg/pix			double	0	0
ra1	Right ascension of first image corner	deg	pos.eq.ra		double	0	0
dec1	Declination of first image corner	deg	pos.eq.dec		double	0	0
ra2	Right ascension of second image corner	deg	pos.eq.ra		double	0	0
dec2	Declination of second image corner	deg	pos.eq.dec		double	0	0
ra3	Right ascension of third image corner	deg	pos.eq.ra		double	0	0
dec3	Declination of third image corner	deg	pos.eq.dec		double	0	0
ra4	Right ascension of fourth image corner	deg	pos.eq.ra		double	0	0
dec4	Declination of fourth image corner	deg	pos.eq.dec		double	0	0
ipac_pub_date	UT date/time of raw product release				char	0	0
ipac_gid	IPAC Group ID				int	0	0
meta_id	Metadata record ID				long	0	1
poly	Bounding polygon
"""
############################
#                          #
#   Meta API Object        #
#                          #
############################
def download_metadata(kind="sci", radec=None, size=None, mcen=None,
                     sql_query=None, colnames=None, **kwargs):
    """  MetaQuery object containing the results of the query in the `metatable` attribute.
    
    Parameters
    ----------
    
    Returns
    -------
    MetaQuery
    """
    mquery = MetaQuery()
    mquery.query(kind=kind, radec=radec, size=size, mcen=mcen,
                 sql_query=sql_query, colnames=colnames, **kwargs)
    return mquery
    
class MetaQuery():

    def query(self, kind="sci", radec=None, size=None, mcen=None,
                  sql_query=None, colnames=None,
                  cookies=None, **kwargs):
        """ Query IRSA to get the metadate needed to know how to access the data themselves.
        
        This method uses `build_query()` and store the downloaded information
        as a panda's DataFrame in the `metatable` attribute.
        The `nentries` attribute will return the number of entry fullfilling the query.

        Parameters
        ----------
        kind: [sting]
             [sci / raw / cal / ref]
        """
        if cookies is None:
            from .io import _load_id_, get_cookie
            cookies = get_cookie(*_load_id_("irsa") )
            
        self.build_query(kind=kind, radec=radec, size=size, mcen=mcen,
                          sql_query=sql_query, colnames=colnames, ct="csv", **kwargs)

        datain = StringIO(
            requests.get( self.query_url, cookies=cookies).content.decode('utf-8')
            )
        self.metatable = read_csv(datain)

        if len(self.metatable.columns) == 1 and "<?xml version" in self.metatable.columns[0]:
            warnings.warn("The query you made failed:" +"\n"+self.query_url+"\n"+"see queriable entries here: %s"%IRSA_DOC)
        elif self.nentries ==0:
            warnings.warn("The query ran successfully, but no data returned. (bad query [check self.query_url]? wrong password [ztfquery.io._load_id_('irsa')] ?)")
            return datain.read()
                              
        
    def build_query(self, kind="sci", radec=None, size=None, mcen=None,
                     sql_query=None, colnames=None, ct="csv", **kwargs):
        """ """
        self.query_prop = dict(kind=kind, radec=
                                   radec, size=size, mcen=mcen, sql_query=sql_query, colnames=colnames, ct=ct, **kwargs)
        self.query_url  = build_query( **self.query_prop )
        
    # ================ #
    #  Properties      #
    # ================ #
    @property
    def datakind(self):
        """ the kind [sci / raw / cal / ref] of data queried """
        if not hasattr(self, "query_prop"):
            raise AttributeError("No `query_prop` loaded yet. Run the build_query() (called by query()) method first. ")
        return self.query_prop['kind']
    
    @property
    def nentries(self):
        """ number of entry returned """
        if not hasattr(self, "metatable"):
            raise AttributeError("No `metatable` loaded yet. Run the query() method first. ")
        return len(self.metatable)
        
############################
#                          #
#   Meta API               #
#                          #
############################
# ================ #
#  Main Function   #
# ================ #
def build_query( kind="sci", radec=None, size=None, mcen=None,
                    sql_query=None, colnames=None, ct="csv"):
    """ 
    Parameters
    ----------
    kind: [sting]
        [sci / raw / cal / ref]

    ct: [string] -optional-
        ouput format
        The query output format is controlled by the ct parameter, which may always be specified, 
        regardless of the query type. The following values are recognized:
        - html  HTML
        - ipac_table IPAC ASCII table format
        - csv Comma-separated-value format
        - tsv Tab-separated-value format

    """
    _test_kind_(kind)
    
    # position
    posquery   = [position_query(radec[0],radec[1], size=size, mcen=mcen)] if radec is not None else []

    colquery   = [columns_query(colnames)]
    if sql_query is None: sql_query=""
    wherequery = [where_query(sql_query.replace(" ","+"))]
    
    query_ = "&".join([l for l in posquery+colquery+wherequery if len(l)>1])
    return SEARCH_BASEURL+"%s?"%kind+query_+"&ct=%s"%ct


def _test_kind_(kind):
    """ """
    if kind not in ["sci", "raw", "cal", "ref"]:
        raise ValueError("Unknown kind '%s'"%kind+" available kinds: sci/raw/cal/ref] ")

# ================ #
#  Query Building  #
# ================ #
def position_query(ra,dec, size=None, mcen=False):
    """ 
    Parameters
    ----------
    ra,dec: [float/str]
         ICRS right ascension and declination in decimal degrees.
         It identifies the point which returned images must contain, or the center of the search region.

    size: [float/str/None] -optional-
        It consists of one or two (comma separated) values in decimal degrees. 
        (With POS=ra,dec)
        The first value is taken to be the full-width of the search region along the east axis at POS, 
        and the second is taken to be the full-height along the north axis. 
        Taken together, POS and SIZE define a convex spherical polygon on the sky with great circle edges - the search region. 
        During a query, this region is compared against the convex spherical polygons formed by connecting 
        the 4 corners of each image in a data-set to determine which images should be returned.

        If only one size value is specified, it is used as both the full-width and full-height.
        Negative sizes are illegal, and a width and height of zero indicate that the search region is a point.

    mcen: [bool] -optional-
        [If the size parameter is specified and non-zero, the mcen parameter is ignored] 

        The mcen parameter indicates that only the most centered image/image set 
        (with respect to POS) should be returned, rather than all images/image sets containing POS. 
    """
    POSITION = ["POS=%s,%s"%(ra,dec)]
    SIZE     = ["SIZE=%s"%size] if size is not None else []
    MCEN     = ["mcen"] if mcen else []
    return "&".join(POSITION+SIZE+MCEN)

def columns_query(colnames):
    """ """
    if colnames is None:
        return ""
    if len(np.atleast_1d(colnames))==1:
        return "COLUMNS=%s"%colnames
    else:
        return "COLUMNS="+",".join(list(np.atleast_1d(colnames)))

# ================ #
#   WHERE QUERY    #
# ================ #
def where_query(generic_sql=None):
    """ 
    The where parameter can be set to a 'SQL WHERE' clause, with some restrictions. 
    [https://en.wikipedia.org/wiki/Where_(SQL)]

    Notably, function calls and sub-queries are not supported. You can use AND, OR, NOT, IN, BETWEEN, LIKE, IS, 
    the usual arithmetic and comparison operators, and literal values.

    Note that the where parameter is required in the absence of POS (a spatial constraint).

    WHERE clauses should be URL encoded [https://en.wikipedia.org/wiki/Query_string#URL_encoding].
    for instance  SPACE is encoded as '+' or "%20".
    If entry must be equal to a string, use `entry='str'` (with the quotes)
        
    Examples:
        get all the science field 600
         - https://irsa.ipac.caltech.edu/ibe/search/ztf/products/sci?WHERE=field=600&ct=html
        get all the science field 600 and having an airmass greater than 2
         - https://irsa.ipac.caltech.edu/ibe/search/ztf/products/sci?WHERE=field=600+AND+airmass>2&ct=html
        get all the science field 600 and having an airmass greater than 2 with a quadran ID been 1 or 3
        - https://irsa.ipac.caltech.edu/ibe/search/ztf/products/sci?WHERE=field=600+AND+airmass>2+AND+qid+IN+(1,3)&ct=html
        
        get observation taken since the 1st of Feb 2018 (julian date 2458150.5) with an airmass > 3
        - https://irsa.ipac.caltech.edu/ibe/search/ztf/products/sci?WHERE=airmass>3+AND+obsjd>2458150.5&ct=html
    
    """
    if generic_sql is None:
        return ""
    else:
        return "WHERE=%s"%generic_sql.replace('"','%27')
    
def get_queriable_field(kind="sci", openurl=False, fulltable=False):
    """ """
    _test_kind_(kind)
    if openurl: fulltable = True
    
    columns = "column_name,description,unit,ucd,utype,datatype,principal,indexed" if fulltable else "column_name"
    url = "https://irsa.ipac.caltech.edu/TAP/sync?query=select+"+columns+ \
      "+from+TAP_SCHEMA.columns++where+table_name=%27ztf.ztf_current_meta_"+kind+"%27+order+by+column_index&format=csv"
      
    if openurl:
        import webbrowser
        webbrowser.open_new_tab( url.replace("format=csv","format=html") )
    else:
        r = requests.get( url )
        data = r.content.decode("utf-8").splitlines()
        if not fulltable: return data[1:]
        return data





""" Main tools to handle ZTF tables (DataFrames) """

import numpy as np
from . import fields


def show_timedf(timedf, ax=None):
    """ """
    if ax is None:
        fig = mpl.figure(figsize=[7,4])
        ax  = fig.add_subplot(111)
    else:
        fig = ax.figure

    

class _ZTFTable_( object ):
    """ """
    def __init__(self, dataframe):
        """ """
        self._data = dataframe
        for mainkey in ["fid","field"]:
            if mainkey not in self.data.columns:
                warning.warn(f"{mainkey} not in the given datatable. Some methods won't work")
        
    # -------- #
    #  GETTER  #
    # -------- #
    def get_filtered(self, field=None, fid=None, grid="both", query=None):
        """ """
        fidflag = True if fid is None else self.data["fid"].isin( np.atleast_1d(fid) )
        fieldflag = True if field is None else self.data["field"].isin( np.atleast_1d(field) )
        gridflag = True if grid is None else self.data["field"].isin( fields.get_grid_field(grid) )
        
        if query is None:
            if fidflag & fieldflag & gridflag is True:
                return self.data
            
            return self.data[fidflag & fieldflag & gridflag]

        return self.data[fidflag & fieldflag & gridflag].query(query)
    
    def get_count(self, entry, normalize=False, query=None, **kwargs):
        """ count the number of time the entry exists 
        For instance, count the number of time a field has been observed:
        self.get_count("field")

        **kwargs goes to get_filtered()
        """
        return self.get_filtered(query=query, **kwargs)[entry].value_counts(normalize=normalize)

    def get_field_average_value(self, entry, groupby="field", method="mean", query=None, **kwargs):
        """ 
        **kwargs goes to get_filtered()
        """
        return getattr(self.get_filtered(query=query, **kwargs).groupby("field"),method)()[entry]
    
    # -------- #
    # PLOTTER  #
    # -------- #
    def show_fields(self, field_val,
                    ax=None, filterprop={},
                    show_ztf_fields=True, grid="main",
                    colorbar=True, cax=None, clabel=" ", 
                    cmap="viridis",
                    vmin=None, vmax=None,  **kwargs):
        """ 
        Parameters
        ----------
        field_val: [dict or pandas.Series]
            Values assocatied to the fields.

        
        """
        from .fields import show_fields
        if type(field_val) is str:        
            if field_val in ["visit","visits","density", "field"]:
                func  = self.get_count
                field_val = "field"
            elif field_val in self.data.columns:
                func  = self.get_field_average_value
            else:
                raise ValueError(f"cannot parse sizeentry {field_val}, could be visits or any data.column")

            field_val = func(field_val, grid=grid, **filterprop)
            
        return show_fields(field_val, ax=ax,
                    show_ztf_fields=show_ztf_fields, grid=grid,
                    colorbar=colorbar, cax=cax, clabel=clabel, 
                    cmap=cmap,
                    vmin=vmin, vmax=vmax,  **kwargs)
    
    def show_gri_fields(self, sizeentry="visits", grid="main", filterprop={}, **kwargs):
        """ """
        if sizeentry in ["visit","visits","density", "field"]:
            func  = self.get_count
            sizeentry = "field"
        elif sizeentry in self.data.columns:
            func  = self.get_field_average_value
        else:
            raise ValueError(f"cannot parse sizeentry {sizeentry}, could be visits or any data.column")
        
        # Data
        fieldsg = func(sizeentry, fid=1, grid=grid, **filterprop)
        fieldsr = func(sizeentry, fid=2, grid=grid, **filterprop)
        fieldsi = func(sizeentry, fid=3, grid=grid, **filterprop)
        # Plot        
        return fields.show_gri_fields(fieldsg, fieldsr, fieldsi, grid=grid, **kwargs)

    def show_evolution(self, value, timekey=None, ax=None, filterprop={}, 
                   groupby=None, groupbymethod="sum", ycoef=1, **kwargs):
        """ """
        import matplotlib.pyplot as mpl
        from matplotlib import dates as mdates
        import pandas
        if ax is None:
            fig = mpl.figure(figsize=[7,4])
            ax  = fig.add_subplot(111)
        else:
            fig = ax.figure

        # - Select the data to plot
        data = self.get_filtered(**filterprop)
        if groupby is not None:
            data = getattr(data.groupby(groupby), groupbymethod)()
            timekey = "index"
        
        #
        # Parsing time
        #
        if timekey is None:
            if "datetime" in data.columns:
                timekey = "datetime"
            elif "obsjd" in data.columns:
                timekey = "obsjd"
            elif "date" in data.columns:
                timekey = "date"
            else:
                raise ValueError(f"cannot automatially find the time key, none of obsjd/datetime/date in data.comluns")
        
        if timekey == "index":
            timearray = pandas.to_datetime(data.index)
        elif timekey == "obsjd":
            from astropy import time
            timearray = time.Time(np.asarray(data[timekey], format="jd").datetime)
        elif timekey in data.columns:
            timearray = pandas.to_datetime(data[timekey])
        else:
            raise ValueError(f"cannot parse the given date key {timekey}")

        # -
        if ycoef is None:
            ycoef = 1
        ax.plot(timearray, data[value]*ycoef, **kwargs)
        # -
    
        locator = mdates.AutoDateLocator()
        formatter = mdates.ConciseDateFormatter(locator)
        ax.xaxis.set_major_locator(locator)
        ax.xaxis.set_major_formatter(formatter)
    
        return fig
    # =============== #
    #  Properties     #
    # =============== #    
    @property
    def data(self):
        """ clean (renamed and reshaped) logs """
        return self._data
    
